# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# type: ignore
import gettext
import pathlib
from typing import Callable

from .translationcatalog import TranslationCatalog
from .types import ITranslation
from .utils import to_language
from .utils import to_locale


CBRA_DOMAIN: str = "cbra"


class Translation(gettext.GNUTranslations, ITranslation):
    """Set up the GNUTranslations context with regard to output charset.

    This translation object will be constructed out of multiple GNUTranslations
    objects by merging their catalogs. It will construct an object for the
    requested language and add a fallback to the default language, if it's
    different from the requested language.
    """
    __module__: str = 'cbra.core.translation'
    default_domain: str = CBRA_DOMAIN
    domain: str
    plural: Callable[[int], int]
    _catalog: TranslationCatalog | None
    __language: str
    __locale: str
    __localedirs: list[str]

    def __init__(
        self,
        language: str,
        domain: str | None = None,
        localedirs: list[str | pathlib.Path] | None = None
    ) -> None:
        super().__init__()
        self.domain = domain or self.default_domain
        self._catalog = None
        self.__language = to_language(language)
        self.__locale = to_locale(language)
        self.__localedirs = [str(x) for x in (localedirs or [])]

        # If a language doesn't have a catalog, use the Germanic default for
        # pluralization: anything except one is pluralized.
        self.plural = lambda n: int(n != 1)

        self._init_catalog()
        for dirname in self.__localedirs:
            self.merge(self.create_gnu_translation(dirname))

        if self._catalog is None:
            self._catalog = TranslationCatalog()

    def _init_catalog(self):
        """Create a base catalog using global translations."""
        self.merge(self.create_gnu_translation(pathlib.Path(__file__).parent.joinpath("locale")))

    def create_gnu_translation(
        self,
        dirname: str,
        use_null_fallback: bool = True
    ) -> gettext.GNUTranslations | gettext.NullTranslations:
        """Return a mergeable :class:`gettext.GNUTranslations` instance.
        A convenience wrapper. By default gettext uses ``fallback=False``.
        Using param `use_null_fallback` to avoid confusion with any other
        references to `fallback`.
        """
        return gettext.translation(
            domain=self.domain,
            localedir=dirname,
            languages=[self.__locale],
            fallback=use_null_fallback
        )

    def merge(self, other: gettext.GNUTranslations | gettext.NullTranslations):
        """Merge another translation into this catalog."""
        if not getattr(other, "_catalog", None):
            assert isinstance(other, gettext.NullTranslations)
            return  # NullTranslations() has no _catalog
        if self._catalog is None:
            # Take plural and _info from first catalog found.
            self.plural = other.plural
            self._info = other._info.copy()
            self._catalog = TranslationCatalog(other)
        else:
            self._catalog.update(other)
        if other._fallback:
            self.add_fallback(other._fallback)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from cbra.types import IAuthorizationContext
from cbra.types import PolicyPrincipal
from cbra.ext.rbac.models import Policy
from cbra.ext.rbac import StaticPermissionsFinder
from cbra.types.policyprincipal import PolicyPrincipal


finder: StaticPermissionsFinder = StaticPermissionsFinder(
    available={
        "foo.create",
        "foo.delete",
        "foo.get",
        "foo.list",
        "foo.update",
        "bar.create",
        "bar.delete",
        "bar.get",
        "bar.list",
        "bar.update",
        "baz.create",
        "baz.delete",
        "baz.get",
        "baz.list",
        "baz.update",
    },
    roles={
        'foo': {'foo.*'}
    }
)


class AuthorizationContext(IAuthorizationContext):

    def __init__(self, email: PolicyPrincipal):
        self.email = email

    def get_principals(self) -> set[PolicyPrincipal]:
        return {self.email}
    

@pytest.mark.asyncio
async def test_granted_by_email():
    policy = Policy()
    policy.add('foo', 'email:foo@example.com')
    ctx = AuthorizationContext(
        email=PolicyPrincipal('email:foo@example.com')
    )
    grants = await policy.grants(finder, ctx)
    assert grants == {'foo.get', 'foo.update', 'foo.list', 'foo.create', 'foo.delete'}


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
import pytest
import pytest_asyncio

from cbra.ext.security import ApplicationKeychain


@pytest_asyncio.fixture # type: ignore
async def keychain() -> ApplicationKeychain:
    return await ApplicationKeychain.setup()


@pytest.mark.asyncio
async def test_encrypt_decrypt_bytes(
    keychain: ApplicationKeychain,
):
    pt = b"foo"
    ct = await keychain.encrypt(pt)
    dt = await keychain.decrypt(ct)
    assert pt == dt


@pytest.mark.asyncio
async def test_encrypt_decrypt_model(
    keychain: ApplicationKeychain,
):
    class Model(pydantic.BaseModel):
        foo: str
        bar: int

    pt = Model(foo='foo', bar=1)
    ct = await keychain.encrypt(pt)
    dt = await keychain.decrypt(ct, parser=lambda _, pt: Model.parse_raw(pt))
    assert pt == dt


@pytest.mark.asyncio
async def test_encrypt_decrypt_str(
    keychain: ApplicationKeychain,
):
    pt = "foo"
    ct = await keychain.encrypt(pt)
    dt = await keychain.decrypt(ct, parser=lambda _, pt: bytes.decode(pt))
    assert pt == dt


@pytest.mark.asyncio
async def test_hmac_str(
    keychain: ApplicationKeychain,
):
    await keychain.create_hmac('foo')


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timedelta
from datetime import timezone

import pydantic
from headless.ext import oauth2


class ResourceServerAccessToken(pydantic.BaseModel):
    """Maintains a single access token for a specific :term:`Resource Server`.
    This storage model may be used to implement shadow tokens, usually in
    for a backend-for-frontend combined with a browser-based application (SPA).
    """
    access_token: str = pydantic.Field(
        default=...
    )

    expires: datetime = pydantic.Field(
        default=...
    )

    grant_id: str = pydantic.Field(
        default=...
    )

    issuer: str = pydantic.Field(
        default=...
    )

    resource: str = pydantic.Field(
        default=...
    )

    @classmethod
    def parse_response(cls, grant_id: str, issuer: str, resource: str, response: oauth2.TokenResponse):
        now = datetime.now(timezone.utc)
        return cls.parse_obj({
            'access_token': str(response.access_token),
            'expires': now + timedelta(seconds=response.expires_in),
            'grant_id': grant_id,
            'issuer': issuer,
            'resource': resource,
        })

    def is_expired(self, now: datetime | None = None) -> bool:
        now = now or datetime.now(timezone.utc)
        return self.expires <= now

    async def refresh(self, client: oauth2.Client) -> None:
        """Refreshes the access token using the given client."""
        raise NotImplementedError

    def __str__(self) -> str:
        return self.access_token


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable

import jinja2
import jinja2.ext
import jinja2.parser
from jinja2 import nodes


class BaseWrapExtension(jinja2.ext.Extension):
    tags: set[str]
    template: str

    def parse(self, parser: jinja2.parser.Parser):
        tag = parser.stream.current.value
        lineno = next(parser.stream).lineno
        args, kwargs = self.parse_args(parser)
        body = parser.parse_statements(tuple(['name:end{}'.format(tag)]), drop_needle=True)

        return nodes.CallBlock(self.call_method('wrap', args, kwargs), [], [], body).set_lineno(lineno)

    def parse_args(self, parser: jinja2.parser.Parser) -> tuple[list[Any], list[Any]]:
        args: list[Any] = []
        kwargs: list[Any] = []
        require_comma = False

        while parser.stream.current.type != 'block_end':
            if require_comma:
                parser.stream.expect('comma')

            if parser.stream.current.type == 'name' and parser.stream.look().type == 'assign':
                key = parser.stream.current.value
                parser.stream.skip(2)
                value = parser.parse_expression()
                kwargs.append(nodes.Keyword(key, value, lineno=value.lineno))
            else:
                if kwargs:
                    parser.fail('Invalid argument syntax for WrapExtension tag',
                                parser.stream.current.lineno)
                args.append(parser.parse_expression())

            require_comma = True

        return args, kwargs

    @jinja2.pass_context
    def wrap(
        self,
        context: dict[str, Any],
        caller: Callable[..., Any],
        template: str | None = None,
        *args: Any,
        **kwargs: Any
    ) -> str:
        t = self.environment.get_template(template or self.template)
        return t.render(dict(context, content=caller(), **kwargs))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cbra.types import ICredentialVerifier
from cbra.types import IRequestPrincipal
from cbra.types import ISubject


class Subject(ISubject):
    sub: str
    principal: IRequestPrincipal
    authenticated: bool = False

    def __init__(
        self,
        *,
        id: str,
        email: str | None,
        principal: IRequestPrincipal
    ):
        self.email = email
        self.sub = id
        self.principal = principal

    async def authenticate(
        self,
        verifier: ICredentialVerifier[Any],
        providers: set[str] | None = None
    ) -> None:
        self.authenticated = await verifier.verify(
            self.principal,
            self.principal.get_credential(),
            providers=providers
        )

    def is_authenticated(self) -> bool:
        return self.authenticated

    def __repr__(self) -> str:
        return f'Subject(id={self.sub})'


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import collections
import inspect
import logging
import uuid
from datetime import datetime
from datetime import timezone
from types import UnionType
from typing import get_args
from typing import get_origin
from typing import Awaitable
from typing import Any
from typing import Callable
from typing import Generic
from typing import TypeVar
from typing import Union
from inspect import Parameter

import fastapi
import fastapi.params
import pydantic

from cbra.types import Abortable
from cbra.types import ETagSet
from cbra.types import IEndpoint
from cbra.types import IntegerPathParameter
from cbra.types import MutableSignature
from cbra.types import PathParameter
from cbra.types import StringPathParameter
from cbra.types import UUIDPathParameter
from .params import IfMatchDestructive
from .params import IfMatchRequestHeader


E = TypeVar('E', bound='IEndpoint')
T = TypeVar('T', bound='RequestHandler[Any]')


class RequestHandler(Generic[E]):
    __module__: str = 'cbra'
    _is_coroutine = asyncio.coroutines._is_coroutine # type: ignore
    _annotations: tuple[type, ...] = (
        fastapi.Request,
        fastapi.Response,
    )
    _class: type[E] | None
    _class_params: set[str]
    _dependencies: dict[str, Parameter] = {}
    _func: Callable[..., Awaitable[Any] | Any]
    _handler_params: set[str]
    _init_params: set[str]
    _injectables: tuple[type, ...] = (
        fastapi.params.Body,
        fastapi.params.Depends,
        fastapi.params.Param,
    )
    _method: str
    _path_types: dict[type, type[PathParameter]] = {
        int: IntegerPathParameter,
        str: StringPathParameter,
        uuid.UUID: UUIDPathParameter
    }
    _signature: inspect.Signature | None
    include_in_schema: bool = True
    logger: logging.Logger = logging.getLogger('cbra.endpoint')
    response_model_by_alias: bool = False
    status_code: int = 200

    @property
    def attname(self) -> str:
        return str.lower(self._method)

    @property
    def endpoint(self) -> type[E]:
        assert self._class is not None
        return self._class

    @property
    def method(self) -> str:
        return str.upper(self._method)

    @property
    def __signature__(self) -> inspect.Signature:
        assert self._signature is not None
        return self._signature

    @property # type: ignore
    def __doc__(self) -> str | None:
        return self._func.__doc__

    def __init__(
        self,
        name: str,
        method: str,
        func: Callable[..., Awaitable[Any] | Any],
        include_in_schema: bool | None = None,
        response_model_by_alias: bool = False
    ):
        self._endpoint_name = name
        self._class = None
        self._class_args = []
        self._class_params = set()
        self._dependencies = {}
        self._handler_params = set()
        self._handler_sig = MutableSignature.fromfunction(func)
        self._init_params = set()
        self._method = method
        self._func = func
        self._func, self._handler_sig = self.validate_handler(self._func, self._handler_sig)
        if include_in_schema is not None:
            self.include_in_schema = include_in_schema
        self.response_model_by_alias = response_model_by_alias
        # Check if the asyncio.iscoroutinefunction() call returns
        # True for this object, since it depends on a private
        # symbol.
        assert asyncio.iscoroutinefunction(self) # nosec

    def clone(self: T) -> T:
        return type(self)(
            self._endpoint_name,
            self.method,
            self._func,
            self.include_in_schema
        )

    def annotate_path(self, cls: Any) -> type[PathParameter]:
        return self._path_types[cls]

    def get_return_annotation(self) -> Any:
        return self._handler_sig.return_annotation

    def is_injectable(self, cls: Any) -> bool:
        """Return a boolean indicating if the class is injectable."""
        cls = get_origin(cls) or cls
        return any([
            inspect.isclass(cls) and issubclass(cls, self._annotations),
            cls in self._annotations
        ])

    def validate_handler(
        self,
        func: Callable[..., Awaitable[Any] | Any],
        signature: MutableSignature
    ) -> tuple[Callable[..., Awaitable[Any] | Any], MutableSignature]:
        return func, signature

    def add_to_class(self, cls: type[IEndpoint]) -> None:
        """Construct an entrypoint for the router and add it to the
        endpoint class.
        """
        self._class = cls # type: ignore
        self._class_sig = MutableSignature.fromfunction(cls)

        # Collect all dependencies and ensure that there are no clashing
        # attribute or parameter names.
        dependencies = self._dependencies
        annotations: dict[str, type] = getattr(cls, '__annotations__') or {}

        # Begin with inspecting the annotations, these might contain automatically
        # injectable objects such as fastapi.Request.
        for attname, annotation in annotations.items():
            if hasattr(cls, attname):
                # Is a dependency or has a default, see below.
                continue
            if not self.can_inject(annotation, 'class'):
                continue
            dependencies[attname] = Parameter(
                kind=Parameter.POSITIONAL_ONLY,
                name=attname,
                annotation=annotation
            )
            self._class_params.add(attname)

        for attname, value in inspect.getmembers(cls):
            if not self.can_inject(value, 'class'):
                continue
            dependencies[attname] = Parameter(
                kind=Parameter.POSITIONAL_OR_KEYWORD,
                name=attname,
                annotation=annotations.get(attname),
                default=value
            )
            self._class_params.add(attname)

        # If the constructor has named parameters, then the __init__
        # was overridden and also might contain dependencies.
        for param in self._class_sig.named:
            if not self.can_inject(param, 'init'):
                raise TypeError("Constructor arguments must be injectable.")
            if param.name in {'request', 'response'}:
                raise ValueError(f'Can not inject {param.name}')
            if param.name in dependencies:
                raise TypeError(
                    f"{cls.__name__} dependency '{param.name}' conflicts "
                    f"with dependency {cls.__name__}.{param.name}."
                )
            dependencies[param.name] = param
            self._init_params.add(param.name)

        # Finally, get the dependencies from the request handler.
        for param in self._handler_sig.named:
            if param.name == 'self' or not self.can_inject(param, 'handler'):
                continue
            if param.name in {'request', 'response'}:
                raise ValueError(f'Can not inject {param.name}')
            if param.name in dependencies:
                raise TypeError(
                    f"{cls.__name__}.{self.attname} dependency "
                    f"'{param.name}' conflicts with {cls.__name__}.{param.name} "
                    "or a constructor argument."
                )

            annotation = param.annotation
            default = param.default

            # Wrap the annotation if it is an injectable path parameter,
            # because we don't want to return 422 like FastAPI does but
            # a proper 404.
            if annotation in self._path_types\
            and not isinstance(default, self._injectables):
                annotation = self.annotate_path(annotation)

            dependencies[param.name] = Parameter(
                kind=(
                    Parameter.POSITIONAL_OR_KEYWORD
                    if param.default != Parameter.empty
                    else Parameter.POSITIONAL_ONLY
                ),
                name=param.name,
                annotation=annotation,
                default=default
            )
            self._handler_params.add(param.name)

        # Sort the parameters so that the order is correct. Also force add
        # some dependencies.
        dependencies['request'] = Parameter(
            kind=Parameter.POSITIONAL_OR_KEYWORD,
            name='request',
            annotation=fastapi.Request
        )
        dependencies['response'] = Parameter(
            kind=Parameter.POSITIONAL_OR_KEYWORD,
            name='response',
            annotation=fastapi.Response
        )
        if getattr(cls, 'versioned', False):
            dependencies['etag'] = Parameter(
                kind=Parameter.POSITIONAL_OR_KEYWORD,
                name='etag',
                annotation=ETagSet,
                default=IfMatchDestructive if self.can_mutate() else IfMatchRequestHeader
            )
            self._class_params.add('etag')
        self._class_params.add('request')
        self._class_params.add('response')

        parameters = list(dependencies.values())
        for i, p in enumerate(parameters):
            parameters[i] = self.preprocess_parameter(p) or p
        parameters, return_annotation = self.preprocess_signature(
            parameters={p.name: p for p in parameters},
            return_annotation=self.get_return_annotation()
        )
        parameters = [
            *[
                p for p in parameters
                if p.kind == Parameter.POSITIONAL_ONLY
                and p.default == Parameter.empty
            ],
            *[
                p for p in parameters
                if p.kind == Parameter.POSITIONAL_ONLY
                and p.default != Parameter.empty
            ],
            *[
                p for p in parameters
                if p.kind == Parameter.POSITIONAL_OR_KEYWORD
                and p.default == Parameter.empty
            ],
            *[
                p for p in parameters
                if p.kind == Parameter.POSITIONAL_OR_KEYWORD
                and p.default != Parameter.empty
            ],
            *[p for p in parameters if p.kind == Parameter.KEYWORD_ONLY],
        ]

        sig = inspect.signature(self.__call__)
        self._dependencies = collections.OrderedDict([
            (param.name, param) for param in parameters
        ])
        self._signature = sig.replace(
            parameters=parameters,
            return_annotation=return_annotation
        )

    def add_to_router(
        self,
        cls: IEndpoint,
        router: fastapi.APIRouter,
        **kwargs: Any
    ) -> None:
        assert self._class is not None
        if self.method == 'OPTIONS':
            kwargs.update({'response_model': None, 'status_code': 200})
        if self._handler_sig.return_annotation == None:
            # Ensure that a proper response code is presented to
            # FastAPI if the handler does not return anything.
            kwargs.update({
                'response_model': None,
                'status_code': 204
            })
        kwargs.setdefault('name', cls.name)
        kwargs.setdefault('response_model_by_alias', cls.response_model_by_alias)
        kwargs.update(getattr(self._func, 'params', {}))
        router.add_api_route(
            endpoint=self,
            methods=[self.method],
            include_in_schema=self._class.include_in_schema\
                and self.include_in_schema,
            **kwargs,
        )

    def can_inject(self, p: Parameter | Any, where: str) -> bool:
        result = any([
            isinstance(p, Parameter)\
                and inspect.isclass(p.annotation) and self.is_injectable(p.annotation),
            isinstance(p, Parameter)\
                and isinstance(p.default, self._injectables),
            isinstance(p, Parameter) and (where=='handler')\
                and p.annotation in (self._path_types),\
            isinstance(p, Parameter) and (where=='handler')\
                and self.is_pydantic_union(p.annotation),
            isinstance(p, Parameter) and (where=='handler')\
                and get_origin(p.annotation) is None\
                and inspect.isclass(p.annotation)\
                and issubclass(p.annotation, pydantic.BaseModel),
            isinstance(p, Parameter) and (where=='handler')\
                and get_origin(p.annotation) == dict,
            not isinstance(p, Parameter)\
                and isinstance(p, self._injectables),
            not isinstance(p, Parameter) and self.is_injectable(p),
            # Request handler parameter that is a union of pydantic model. For POST, PUT,
            # PATCH and DELETE handlers, these are injected as request body.
            isinstance(p, Parameter) and (where == 'handler')\
                and get_origin(p.annotation) == UnionType\
                and all([isinstance(x, pydantic.main.ModelMetaclass) for x in get_args(p.annotation)])
        ])
        return result
    
    def can_write(self) -> bool:
        return self.method in {'POST', 'PUT', 'PATCH', 'DELETE'}
    
    def can_mutate(self) -> bool:
        return self.method in {'PUT', 'PATCH'}

    def is_pydantic_union(self, obj: Any):
        origin = get_origin(obj) # type: ignore
        if origin != Union:
            return False
        return all([
            isinstance(x, pydantic.main.ModelMetaclass)
            for x in get_args(obj)
        ])

    def preprocess_parameter(self, p: Parameter) -> Parameter | None:
        """Hook to modify a parameter just before it is added to the
        new signature. It is expected to return a modified
        :class:`inspect.Parameter` instance.
        """
        return p

    def preprocess_signature(
        self,
        parameters: dict[str, Parameter],
        return_annotation: Any
    ) -> tuple[list[Parameter], Any]:
        return list(parameters.values()), return_annotation

    async def preprocess_value(self, name: str, value: Any) -> Any:
        if name == 'etag' and value:
            value = ETagSet.toset(value)
        return value

    async def process_response(
        self,
        endpoint: IEndpoint,
        response: fastapi.Response | pydantic.BaseModel | None
    ) -> fastapi.Response:
        if response is None:
            response = fastapi.Response(status_code=204)
        elif isinstance(response, pydantic.BaseModel):
            response = fastapi.responses.Response(
                headers={'Content-Type': "application/json"},
                status_code=self.status_code,
                content=response.json(
                    by_alias=endpoint.response_model_by_alias,
                    exclude=endpoint.response_model_exclude,
                    exclude_none=endpoint.response_model_exclude_none,
                    indent=2,
                )
            )
        elif isinstance(response, dict):
            response = fastapi.responses.JSONResponse(
                status_code=self.status_code,
                content=response
            )
        assert isinstance(response, fastapi.Response), type(response) # nosec
        response.headers.update(endpoint.get_success_headers(response))
        return response

    async def __call__(self, **params: Any) -> Any:
        try:
            return await self._run_handler(**params)
        except Abortable as exc:
            return await exc.as_response()

    async def _run_handler(self, **params: Any) -> Any:
        # Construct the init arguments, instance attributes and handler
        # arguments from the known parameters.
        assert self._class is not None
        attrs: dict[str, Any] = {}
        init: dict[str, Any] = {}
        kwargs: dict[str, Any] = {}

        for param in self._dependencies.values():
            value = await self.preprocess_value(param.name, params.pop(param.name))

            # PathParameter instances expose a clean() method
            # that immetialy cause the endpoint to return 404
            # on validation failure.
            if inspect.isclass(param.annotation)\
            and issubclass(param.annotation, PathParameter):
                value = param.annotation.clean(value)

            if param.name in self._init_params:
                init[param.name] = value
            elif param.name in self._class_params:
                attrs[param.name] = value
            elif param.name in self._handler_params:
                kwargs[param.name] = value

        if params:
            raise TypeError("Received unknown arguments.")

        # Initialize the endpoint and set its attributes, proceed
        # to invoke the handler.
        endpoint: IEndpoint = self._class(**init)
        endpoint.timestamp = datetime.now(timezone.utc)
        endpoint.__dict__.update(attrs)
        response = await self.process_response(
            endpoint,
            await endpoint.run_handler(self._func, **kwargs)
        )

        # TODO: Make this configurable on the endpoint/resource.
        response.headers.setdefault('Cache-Control', 'no-cache, no-store, must-revalidate')
        response.headers.setdefault('Expires', '0')
        response.headers.setdefault('Permissions-Policy', 'interest-cohort=()')
        response.headers.setdefault('Pragma', 'no-cache')
        response.headers.setdefault('Referrer-Policy', 'no-referrer')
        response.headers.setdefault('X-Content-Type-Options', 'nosniff')
        response.headers.setdefault('X-Frame-Options', 'DENY')
        response.headers.setdefault('X-DNS-Prefetch-Control', 'off')
        response.headers.setdefault('X-XSS-Protection', '0')

        if endpoint.session.is_dirty():
            self.logger.debug("Setting session cookie")
            await endpoint.session.add_to_response(response)

        if endpoint.csrf_protect:
            await endpoint.session
            response.set_cookie(
                key='csrftoken',
                value=endpoint.session.csrf,
                httponly=True,
                max_age=86400,
                samesite='strict',
                secure=True,
            )

        # Copy the headers from the endpoint response.
        response.raw_headers = [
            *response.raw_headers,
            *endpoint.response.raw_headers
        ]

        return response


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Iterable

import cbra.core as cbra
from cbra.core.conf import settings
from cbra.core.iam import NullAuthorizationContext
from cbra.types import IAuthorizationContext
from ..params import RequestAccessToken
from ..types import BearerTokenException
from ..resourceserversubjectresolver import ResourceServerSubjectResolver
from .resourceserverauthorizationcontext import ResourceServerAuthorizationContext
from .types import Scope


class ResourceServerResource(cbra.Resource, model=None): # type: ignore
    __abstract__: bool = True
    __module__: str = 'cbra.ext.oauth2.resource'
    principal: RequestAccessToken # type: ignore
    required_scope: Iterable[Scope] = set()
    subjects: ResourceServerSubjectResolver

    def get_trusted_issuers(self) -> set[str]:
        return settings.OAUTH2_TRUSTED_ISSUERS

    async def authenticate(self) -> None:
        self.ctx = await self.setup_context()
    
    async def setup_context(self) -> IAuthorizationContext:
        if self.principal.is_anonymous():
            assert self.request.client is not None
            return NullAuthorizationContext(
                remote_host=self.request.client.host
            )
        await self.verify_token()
        await self.verify_scope()
        return ResourceServerAuthorizationContext(
            self.request,
            await self.subjects.resolve(self.principal)
        )

    async def verify_scope(self):
        granted = self.principal.get_scope()
        missing: set[str] = set()
        for scope in self.required_scope:
            missing.update(scope.missing(granted))
        if missing:
            raise BearerTokenException(
                scheme='Bearer',
                error='insufficient_scope',
                error_description=(
                    "The access token was not granted the required scope for this "
                    "resource."
                ),
                scope=missing,
                status_code=403
            )
    
    async def verify_token(self):
        await self.principal.verify(
            request=self.request,
            trusted_issuers=self.get_trusted_issuers()
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
.. _ref-guides-persistence

=====================================
Storing and retrieving data with CBRA
=====================================
"""
from typing import Any
from typing import Protocol
from typing import TypeVar


I = TypeVar('I')
M = TypeVar('M')


class IModelRepository(Protocol[M]):
    model: type[M]

    async def allocate(self, model: type[Any], n: int = 1) -> int | list[int]:
        raise NotImplementedError

    def construct_key(self, model: Any) -> Any:
        raise NotImplementedError

    def restore(self, model: type[Any], entity: Any) -> Any:
        raise NotImplementedError

    async def get(self, identity: Any) -> M | None:
        raise NotImplementedError

    async def persist(self, instance: M) -> M:
        """Persist the model. When invoking this method, it is assumed that
        all necessary checks have been performed, so no validation is
        performed here.
        """
        raise NotImplementedError


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from typing import Any

import fastapi
from starlette.datastructures import URL

from cbra.types import BaseModel
from cbra.types import RequestPrincipal
from cbra.types import IEmailSender
from cbra.types import IRoutable
from ..iam import AuthorizationContextFactory
from ..ioc import instance
from ..messagepublisher import MessagePublisher
from ..params import ApplicationEmailSender
from ..sessions import RequestSession
from .const import CURSOR_PARAM_NAME
from .iresource import IResource
from .resourcemodel import ResourceModel
from .resourcetype import ResourceType


class Resource(IResource, metaclass=ResourceType):
    __abstract__: bool = True
    __actions__: list[type[IRoutable]] = []
    __module__: str = 'cbra.core'
    context_factory: AuthorizationContextFactory = AuthorizationContextFactory.depends()
    cursor_param_name: str = CURSOR_PARAM_NAME
    email: IEmailSender = ApplicationEmailSender
    max_limit: int = 100
    model: type[ResourceModel]
    path_prefix: str | None = None
    principal: RequestPrincipal = RequestPrincipal.depends()
    publisher: MessagePublisher = instance('MessagePublisher')
    response_model: BaseModel
    response_model_by_alias: bool = True
    session: RequestSession = RequestSession.depends()
    timestamp: datetime

    def __init_subclass__(cls, model: type[ResourceModel] | None):
        if model is not None:
            cls.model = model

    @classmethod
    def add_to_router(cls, router: fastapi.FastAPI, **kwargs: Any) -> None:
        path = str.rstrip(kwargs.get('path') or '', '/')
        if cls.path_prefix:
            if not str.startswith(cls.path_prefix, '/'):
                raise ValueError(f'{cls.__name__}.path_prefix must start with a slash.')
            kwargs['path'] = f'{path}{cls.path_prefix}'
        for action in cls.__actions__:
            kwargs.setdefault('response_model_by_alias', cls.response_model_by_alias)
            action.add_to_router(cls, router, **kwargs)

    def adapt(self, obj: Any) -> Any:
        """Adapt an object of any kind to a :class:`ResourceModel` instance.
        The default implementation simply returns the object.
        """
        return obj
    
    def parse_obj(self, obj: Any) -> Any:
        return self.model.parse_obj(obj)
    
    def reverse(self, action: str, params: dict[str, Any]) -> URL:
        query: dict[str, str] = {}
        cursor = params.pop(self.cursor_param_name, None)
        url = self.request.url_for(f'{self.path_name}.{action}')
        if cursor is not None:
            query[self.cursor_param_name] = cursor
        if query:
            url = url.include_query_params(**query)
        return url


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Iterable

import fastapi

from .abortable import Abortable


class ValidationFailure(Abortable):
    """Indicates the definitive validation failure of the input data for a given
    operation.
    """
    __module__: str = 'cbra.types'
    status_code: int = 422
    errors: dict[str, set[str]]

    def __init__(self, errors: Iterable[tuple[str, Iterable[str]]]) -> None:
        self.errors = {}
        for name, codes in errors:
            if name not in self.errors:
                self.errors[name] = set()
            self.errors[name].update(codes)

    async def as_response(self) -> fastapi.responses.JSONResponse:
        return fastapi.responses.JSONResponse(
            status_code=self.status_code,
            content={
                'errors': {k: list(v) for k, v in self.errors.items()}
            }
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import json
from typing import Any
from typing import Callable
from typing import Iterable
from typing import TypeVar

import pydantic
from google.cloud.datastore import Client

import cbra.core as cbra
from cbra.ext.security import ApplicationKeychain
from cbra.types import PersistedModel
from cbra.types import ModelMetadata
from cbra.types import IPolymorphicCursor
from .basemodelrepository import BaseModelRepository
from .polymoprhicdatastorecursor import PolymorphicDatatoreCursor
from .types import IDatastoreKey
from .types import IDatastoreEntity

T = TypeVar('T', bound=PersistedModel)
M = TypeVar('M', bound=pydantic.BaseModel)


class PolymorphicDatastoreRepository(BaseModelRepository):
    __module__: str = 'cbra.ext.google'
    keychain: ApplicationKeychain

    def __init__(
        self,
        client: Client | Any = cbra.inject('GoogleDatastoreClient'),
        keychain: ApplicationKeychain = ApplicationKeychain.depends()
    ):
        self.client = client
        self.keychain = keychain

    def model_to_entity(self, obj: PersistedModel) -> IDatastoreEntity:
        """Convert a :class:`cbra.types.PersistedModel` to a valid
        :class:`~cbra.ext.google.types.IDatastoreEntity` implementation.
        """
        pk = self.inspect.get_primary_key_field(obj)
        if pk is None:
            raise TypeError(f'Model {type(obj).__name__} does not have a primary key.')
        key = self.model_key(obj)
        data = obj.json(exclude={pk.name})
        return self.entity_factory(key, _metadata=obj.__metadata__.dict(), **json.loads(data))

    async def auto_increment(self, obj: type[PersistedModel] | PersistedModel) -> int:
        """Return an auto incrementing integer to be used as a technical
        primary key.
        """
        return await self.allocate(self.inspect.get_entity_name(obj))

    async def find(
        self,
        cls: type[T],
        filters: list[tuple[str, str, Any]],
        ordering: list[str] | None = None
    ) -> T | None:
        """Find an entity using the given filters."""
        q = self.query(kind=cls.__name__)
        if ordering:
            q.order = ordering
        for attname, op, value in filters:
            q.add_filter(attname, op, value)
        obj: IDatastoreEntity | None = None
        for entity in (await self.run_in_executor(q.fetch)):
            if obj is not None:
                raise ValueError("Multiple objects returned")
            obj = entity
            break
        if obj is not None:
            field = self.inspect.get_primary_key_field(cls)
            assert field is not None
            return cls.parse_obj({**dict(obj), field.name: obj.key.id or obj.key.name})

    async def get(self, cls: type[T], pk: Any) -> None | T:
        """Lookup an entity by its primary key."""
        entity = await self.get_entity_by_key(self.model_key(cls, pk))
        if entity is None:
            return None
        field = self.inspect.get_primary_key_field(cls)
        assert field is not None
        return cls.parse_obj({**dict(entity), field.name: pk})


    async def get_metadata(self, key: IDatastoreKey) -> ModelMetadata | None:
        """Return the metadata for the given key, or ``None`` if there is no
        metadata.
        """
        entity = dict((await self.get_entity_by_key(key)) or {}) # type: ignore
        if entity:
            return ModelMetadata.parse_obj(entity['_metadata'])

    async def list(
        self,
        cls: type[T],
        ordering: list[str] | None = None,
        limit: int = 100,
        token: str | None = None,
        adapter: Callable[[T], M] | None = None
    ) -> IPolymorphicCursor[M]:
        query = self.query(kind=cls.__name__)
        if ordering:
            query.order = ordering
        result = await self.run_in_executor(
            functools.partial(
                query.fetch,
                start_cursor=token,
                limit=limit
            )
        )
        return PolymorphicDatatoreCursor(cls, [x for x in result], result.next_page_token, adapter=adapter) # type: ignore

    async def persist(self, obj: T) -> T:
        if not isinstance(obj, PersistedModel):
            raise TypeError(type(obj).__name__)
        entity = self.model_to_entity(obj)
        self.inspect.check_metadata(
            old=await self.get_metadata(entity.key),
            new=obj.__metadata__
        )
        await self.put(entity)
        return obj

    async def persist_many(self, objects: Iterable[T]) -> Iterable[T]:
        await self.put_multi([self.model_to_entity(x) for x in objects])
        return objects


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
from headless.ext.oauth2.models import ClaimSet

from cbra.types import NotFound
from cbra.types import SessionRequestPrincipal
from .endpoints import AuthorizationServerEndpoint
from .models import AuthorizationRequest
from .models import CurrentAuthorizationRequest
from .models import PatchAuthorizationRequest
from .models import PatchAuthorizationResponse
from .types import AuthorizationRequestIdentifier
from .types import ClientIdentifier
from .types import RedirectURI
from .types import PrincipalIdentifier


class AuthorizationRequestEndpoint(AuthorizationServerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    name: str = 'oauth2.authorize'
    principal: SessionRequestPrincipal # type: ignore
    path: str = '/requests'
    request_id: AuthorizationRequestIdentifier | None = fastapi.Cookie(
        default=None,
        title="Request ID",
        alias='oauth2.request',
        description=(
            "The authorization request identifier. This cookie is set by the "
            "authorization endpoint in the case that the resource owner "
            "needs to perform a certain action."
        )
    )
    summary: str = 'Authorization Request'

    async def get_authorization_request(
        self,
        request_id: AuthorizationRequestIdentifier | None
    ) -> AuthorizationRequest:
        if not request_id:
            raise NotFound
        await self.session
        request = await self.storage.get(AuthorizationRequest, request_id)
        if request is None:
            self.logger.debug("Authorization request does not exist")
            raise NotFound
        if request.session_id != self.session.id:
            self.logger.debug(
                "Authorization request lookup by unknown session (expected: %s, actual: %s)",
                request.session_id, self.session.id
            )
            raise NotFound
        return request

    async def get(
        self,
        query_request_id: AuthorizationRequestIdentifier | None = fastapi.Query(
            default=None,
            title="Request ID",
            alias="request_id",
            description=("The authorization request identifier.")
        )
    ) -> CurrentAuthorizationRequest:
        request = await self.get_authorization_request(query_request_id or self.request_id)
        requires: set[str] = set()
        if self.is_authenticated():
            requires = request.needs(await self.get_subject()) # type: ignore
        return CurrentAuthorizationRequest(
            request_id=request.id,
            client=request.client_info,
            consent=request.consent,
            email=request.email,
            scope=request.scope,
            id_token=ClaimSet.parse_obj({'sub': '0', **request.id_token}),
            requires=requires
        )
    
    async def patch(
        self,
        dto: PatchAuthorizationRequest,
        query_request_id: AuthorizationRequestIdentifier | None = fastapi.Query(
            default=None,
            title="Request ID",
            alias="request_id",
            description=("The authorization request identifier.")
        )
    ) -> PatchAuthorizationResponse:
        request = await self.get_authorization_request(query_request_id or self.request_id)
        url = None
        if dto.deny:
            redirect_uri = RedirectURI(request.redirect_uri)
            url = redirect_uri.redirect(error='access_denied')
        elif dto.email:
            principal = await self.storage.fetch(PrincipalIdentifier(dto.email))
            if principal is None or not principal.is_owned_by(self.session.uid):
                raise NotFound
            request.set_email(str(principal), prompted=True)
            url = RedirectURI(request.get_authorize_url(self.request))
        elif dto.logout:
            await self.session
            client = await self.storage.fetch(ClientIdentifier(request.client_id))
            self.session.logout()
            request = request.clone()
            await request.load(client, self.storage, self.session.id)
            url = RedirectURI(request.get_authorize_url(self.request))
        else:
            raise NotImplementedError
        assert url is not None
        await request.persist(self.storage)
        return PatchAuthorizationResponse(next=url)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from canonical import EmailAddress


class EmailChallengeRequest(pydantic.BaseModel):
    email: EmailAddress = pydantic.Field(
        default=...,
        title="Email address",
        description=(
            "The email address to send a verification code to in order to verify "
            "ownership."
        ),
        max_length=320
    )

    def is_request(self) -> bool:
        return True

    def is_solution(self) -> bool:
        return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from types import NotImplementedType

import pydantic
from cbra.types import IDependant
from cbra.types import IVerifier
from cbra.types import Request


class WebhookEnvelope(IDependant):
    __module__: str = 'cbra.ext.webhooks'
    event_name: str | NotImplementedType = NotImplementedType
    logger: logging.Logger = logging.getLogger('cbra.endpoint.webhooks')

    def get_message(self) -> pydantic.BaseModel:
        raise NotImplementedError

    async def verify(
        self,
        request: Request,
        verifier: IVerifier
    ) -> bool:
        """Subclasses must override this method to verify the authenticity "
        and integrity of the webhook message.
        """
        raise NotImplementedError("Subclasses must override this method.")


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import gettext
from typing import Callable


class TranslationCatalog:
    __module__: str = 'cbra.core.translation'
    _catalogs: list[dict[str, str]]
    _plurals: list[Callable[[int], int]]

    def __init__(self, trans: gettext.NullTranslations | None = None):
        self._catalogs = [trans._catalog.copy()] if trans else [{}] # type: ignore
        self._plurals = [trans.plural] if trans else [lambda n: int(n != 1)] # type: ignore

    def __getitem__(self, key: str) -> str:
        for cat in self._catalogs:
            try:
                return cat[key]
            except KeyError:
                pass
        raise KeyError(key)

    def __setitem__(self, key: str, value: str):
        self._catalogs[0][key] = value

    def __contains__(self, key: str):
        return any(key in cat for cat in self._catalogs)

    def items(self):
        for cat in self._catalogs:
            yield from cat.items()

    def keys(self):
        for cat in self._catalogs:
            yield from cat.keys()

    def update(self, trans: gettext.NullTranslations):
        # Merge if plural function is the same, else prepend.
        for cat, plural in zip(self._catalogs, self._plurals):
            if trans.plural.__code__ == plural.__code__: # type: ignore
                cat.update(trans._catalog) # type: ignore
                break
        else:
            self._catalogs.insert(0, trans._catalog.copy()) # type: ignore
            self._plurals.insert(0, trans.plural) # type: ignore

    def get(self, key: str, default: str | None = None) -> str | None:
        missing = object()
        for cat in self._catalogs:
            result = cat.get(key, missing)
            if result is not missing:
                return result # type: ignore
        return default

    def plural(self, msgid: str, num: int):
        for cat, plural in zip(self._catalogs, self._plurals):
            tmsg = cat.get((msgid, plural(num))) # type: ignore
            if tmsg is not None:
                return tmsg
        raise KeyError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Any

from canonical import EmailAddress

import cbra.core as cbra
from cbra.core.conf import settings
from cbra.types import IPolymorphicRepository
from cbra.types import PersistedModel
from cbra.types import Ratelimited
from cbra.types import IEmailSender
from ..models import EmailChallenge
from ..types import IEmailChallenge
from .models import EmailChallengeRequest
from .models import EmailChallengeResponse
from .models import EmailChallengeSolutionRequest
from .models import EmailChallengeSolutionResponse


class EmailChallengeEndpoint(cbra.Endpoint):
    email_template: str = 'email-verification.html.j2'
    email: IEmailSender = cbra.instance('EmailSender')
    model: type[IEmailChallenge] = EmailChallenge
    require_authentication: bool = False
    response_model_by_alias: bool = True

    def get_email_sender(self) -> str:
        assert settings.EMAIL_DEFAULT_SENDER is not None
        return settings.EMAIL_DEFAULT_SENDER

    async def post(
        self,
        dto: EmailChallengeSolutionRequest | EmailChallengeRequest,
        repo: IPolymorphicRepository = cbra.instance('PolymorphicRepository')
    ) -> EmailChallengeResponse | EmailChallengeSolutionResponse:
        """Provides an interface to create and solve a challenge to
        verify the ownership of an email address.

        To create a challenge, issue a `POST` request containing a JSON
        object of type `EmailChallengeRequest` as the request body. The
        response contains a JSON object of type `EmailChallengeResponse`.
        The client must retain the included `challengeId`. An email is
        sent to the email address (provided with the `email` parameter
        in the request body) containing a verification code.
        """
        handler = self.handle_create if dto.is_request() else self.handle_solve
        return await handler(dto, repo) # type: ignore

    async def handle_create(
        self,
        dto: EmailChallengeRequest,
        repo: IPolymorphicRepository
    ) -> EmailChallengeResponse:
        # Check if an existing challenge exists for this emailaddress,
        # and if it is not old enough, refuse to create a new challenge.
        challenge = await self.retrieve_challenge_by_email(repo, dto.email)
        if challenge and challenge.age() < 60:
            raise Ratelimited
        challenge = await self.create_challenge(dto)
        await self.persist_challenge(repo, challenge)
        await self.on_challenge_created(challenge)
        return EmailChallengeResponse(
            challengeId=challenge.get_challenge_id()
        )

    async def handle_solve(
        self,
        dto: EmailChallengeSolutionRequest,
        repo: IPolymorphicRepository
    ) -> EmailChallengeSolutionResponse:
        challenge = await self.retrieve_challenge(repo, dto.challenge_id)
        if not challenge or challenge.is_blocked():
            return EmailChallengeSolutionResponse(
                success=False,
                blocked=True
            )
        is_solved = challenge.verify(dto.code)
        if is_solved:
            await self.on_challenge_solved(challenge, dto)
        await self.persist_challenge(repo, challenge)
        return EmailChallengeSolutionResponse(
            success=is_solved,
            blocked=challenge.is_blocked()
        )

    async def create_challenge(self, dto: Any) -> IEmailChallenge:
        return self.model.new(dto.email)

    async def retrieve_challenge(
        self,
        repo: IPolymorphicRepository,
        challenge_id: str
    ) -> IEmailChallenge:
        return await repo.get(self.model, challenge_id) # type: ignore
    
    async def retrieve_challenge_by_email(
        self,
        repo: IPolymorphicRepository,
        email: str
    ) -> IEmailChallenge | None:
        return await repo.find(self.model, [('email', '=', email)], ['-requested']) # type: ignore

    async def persist_challenge(
        self,
        repo: IPolymorphicRepository,
        challenge: IEmailChallenge
    ) -> None:
        await repo.persist(cast(PersistedModel, challenge))

    async def on_challenge_created(self, challenge: IEmailChallenge) -> None:
        pass

    async def on_challenge_requested(self, email: EmailAddress) -> None:
        raise NotImplementedError
    
    async def on_challenge_solved(self, challenge: Any, dto: Any) -> None:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
from typing import Any
from typing import TypeVar

from google.cloud.datastore import Client

import cbra.core as cbra
from cbra.ext.security import ApplicationKeychain
from cbra.types import PersistedModel
from cbra.types import ModelInspector
from cbra.types import ModelMetadata
from .basedatastorerepository import BaseDatastoreRepository
from .types import IDatastoreEntity
from .types import IDatastoreKey


T = TypeVar('T', bound=PersistedModel)


class BaseModelRepository(BaseDatastoreRepository):
    __module__: str = 'cbra.ext.google'
    inspect: ModelInspector = ModelInspector()

    def __init__(
        self,
        client: Client | Any = cbra.inject('GoogleDatastoreClient'),
        keychain: ApplicationKeychain = ApplicationKeychain.depends()
    ):
        self.client = client
        self.keychain = keychain

    def check_metadata(
        self,
        old: ModelMetadata | None,
        new: ModelMetadata
    ):
        self.inspect.check_metadata(old=old, new=new)

    def get_entity_name(self, model: type[PersistedModel] | PersistedModel) -> str:
        return self.inspect.get_entity_name(model)

    def get_primary_key(self, obj: PersistedModel) -> Any:
        """Return the primary key value of the object."""
        field = self.inspect.get_primary_key_field(obj)
        if field is None:
            raise TypeError(f'Model {type(obj).__name__} does not have a primary key.')
        return getattr(obj, field.name)

    def model_key(
        self,
        obj: type[PersistedModel] | PersistedModel,
        pk: Any | None = None,
        parent: PersistedModel | IDatastoreKey | None = None
    ) -> IDatastoreKey:
        """Return a :class:`cbra.ext.google.types.IDatastoreKey` instance
        based on the primary key of the model.
        """
        if isinstance(parent, PersistedModel):
            parent = self.model_key(parent)
        if not pk and isinstance(obj, PersistedModel):
            #if not isinstance(obj, PersistedModel):
            #    raise TypeError("The 'pk' parameter can not be None.")
            pk = self.get_primary_key(obj)
        return self.key(self.inspect.get_entity_name(obj), pk, parent=parent)

    def model_to_entity(
        self,
        obj: PersistedModel,
        key: IDatastoreKey | None = None,
        parent: PersistedModel | None = None
    ) -> IDatastoreEntity:
        """Convert a :class:`cbra.types.PersistedModel` to a valid
        :class:`~cbra.ext.google.types.IDatastoreEntity` implementation.
        """
        pk = self.inspect.get_primary_key_field(obj)
        if pk is None:
            raise TypeError(f'Model {type(obj).__name__} does not have a primary key.')
        key = key or self.model_key(obj, parent=parent)
        data = obj.json(exclude={pk.name})
        return self.entity_factory(key, _metadata=obj.__metadata__.dict(), **json.loads(data))

    def restore(self, model: type[T], entity: IDatastoreEntity) -> T:
        field = self.inspect.get_primary_key_field(model)
        assert field is not None
        obj = model.parse_obj({
            **dict(entity),
            field.name: entity.key.id or entity.key.name,
        })
        obj.__metadata__ = ModelMetadata.parse_obj(dict(entity['_metadata']))
        return obj

    async def get_metadata(self, key: IDatastoreKey) -> ModelMetadata | None:
        """Return the metadata for the given key, or ``None`` if there is no
        metadata.
        """
        entity = dict((await self.get_entity_by_key(key)) or {}) # type: ignore
        if entity:
            return ModelMetadata.parse_obj(entity['_metadata'])

    async def persist_model(self, obj: T, parent: PersistedModel | None = None, force: bool = False) -> T:
        if not isinstance(obj, PersistedModel):
            raise TypeError(type(obj).__name__)
        async with self.transaction():
            if not force:
                self.check_metadata(
                    old=await self.get_metadata(self.model_key(obj, parent=parent)),
                    new=obj.__metadata__
                )
                obj.__metadata__.generation += 1
            entity = self.model_to_entity(obj, parent=parent)
            await self.put(entity)
        return obj


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import hashlib
from typing import Any
from typing import TypeVar

import pydantic

from .rfc9068accesstoken import RFC9068AccessToken


T = TypeVar('T', bound='IssuedAccessToken')


class IssuedAccessToken(pydantic.BaseModel):
    claims: dict[str, Any] = {}
    client_id: str
    encrypted: bool = False
    expires: datetime.datetime
    issuer: str
    scope: list[str]
    sub: int | str
    token_hash: str

    @classmethod
    def parse_rfc9068(
        cls: type[T],
        at: RFC9068AccessToken,
        claims: dict[str, Any],
        signed_token: str,
        scope: list[str],
        sub: int | str
    ) -> T:
        return cls.parse_obj({
            'claims': claims,
            'client_id': at.client_id,
            'expires': at.get_expire_date(),
            'issuer': at.iss,
            'scope': scope,
            'sub': sub,
            'token_hash': hashlib.md5(str.encode(signed_token, 'utf-8')).hexdigest()
        })
    
    def is_expired(self, now: datetime.datetime | None = None) -> bool:
        now = now or datetime.datetime.now(datetime.timezone.utc)
        return self.expires < now


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Tests that a model field that is declared as optional is also allowed
on the response model.
"""
import collections

import pydantic
import pytest
import pytest_asyncio
from headless.core import httpx

import cbra.core as cbra


QueryResult = collections.namedtuple('QueryResult', ['token', 'items'])


class AliasModel(cbra.ResourceModel):
    id: int = pydantic.Field(..., primary_key=True)
    value: str | None = pydantic.Field(..., alias='aliasedValue')


class AliasedResource(cbra.Resource, cbra.Collection, model=AliasModel):
    response_model_by_alias: bool = True

    async def retrieve(self, aliasmodel_id: int) -> AliasModel:
        return AliasModel(id=1, aliasedValue=None)

    async def filter(self, *args, **kwargs):
        return QueryResult('', [AliasModel(id=1, aliasedValue=None)])


@pytest_asyncio.fixture(scope='session') # type: ignore
async def client():
    app = cbra.Application()
    app.add(AliasedResource)
    async with httpx.Client(base_url='https://cbra', app=app) as client:
        yield client


@pytest.mark.asyncio
async def test_retrieve_response_model_with_aliases(
    client: httpx.Client
):
    response = await client.get(url='/aliasmodels/1')
    assert response.status_code == 200, response.content


@pytest.mark.asyncio
async def test_list_response_model_with_aliases(
    client: httpx.Client
):
    response = await client.get(url='/aliasmodels')
    assert response.status_code == 200, response.content


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import enum


class CredentialScope(str, enum.Enum):
    #: Global, application-level scope. Use this type for credentials that
    #: can not be provided through configuration (such as environment
    #: variables), for example credentials obtained from an authorization
    #: server.
    GLOBAL = 'GLOBAL'

    #: The credentials are scoped to a single account, or similar (
    #: such as a Subscription in Microsoft Azure or an Organization on
    #: Google Cloud Platform).
    ACCOUNT = 'ACCOUNT'

    #: The credentials are scoped to a single subject.
    SUBJECT = 'SUBJECT'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.types import Abortable
from cbra.types import Request
from ..types import RedirectURI


class AuthorizationException(Abortable):

    def __init__(self, error: str, redirect_uri: RedirectURI) -> None:
        self.error = error
        self.redirect_uri = redirect_uri

    async def as_response(self) -> fastapi.Response:
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=self.redirect_uri.redirect(error=self.error)
        )


def get(
    request: Request,
    error: str | None = fastapi.Query(
        default=None,
        title="Error",
        description="A code describing an error condition."
    ),
    redirect_uri: str = fastapi.Cookie(
        default=...,
        title="Redirect URI",
        alias='bff.redirect_uri',
        description=(
            "The URI to redirect the user-agent to after completing the "
            "login."
        )
    )
) -> AuthorizationException | None:
    if error is not None:
        if redirect_uri.startswith('/'):
            redirect_uri = f'{request.url.scheme}://{request.url.netloc}{redirect_uri}'
        return AuthorizationException(error, RedirectURI(redirect_uri))
    

FrontendError: AuthorizationException | None = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.


class ICredential:
    """Used in conjunction with a :class:`IRequestPrincipal` implementation
    to establish the identity of a subject.
    """
    __module__: str = 'cbra.types'

    def is_verified(self) -> bool:
        """Some credentials are verified simply by existing."""
        return False


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeVar

from cbra.types import Forbidden
from cbra.types import NotFound
from .resourcemodel import ResourceModel
from .persister import Persister
from .viewer import Viewer


T = TypeVar('T', bound=ResourceModel)


class Update(Persister, Viewer):
    __module__: str = 'cbra.core'

    async def can_update(
        self,
        old: T,
        new: T
    ) -> bool:
        """Return a boolean indicating if the resource may be updated
        based on semantics. If this method returns ``False``, then
        the request is rejected with the ``403`` status code.
        """
        return True

    async def update(self, resource: T) -> T:
        old = await self.get_object()
        if old is None:
            raise NotFound
        new = await self.perform_update(
            old=old,
            new=resource
        )
        if not await self.can_update(old, new):
            raise Forbidden
        await self.persist(new, create=False)
        return self.adapt(new)

    async def perform_update(
        self,
        old: T,
        new: T
    ) -> T:
        return old.merge(new)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools

from cbra.ext.security import ApplicationKeychain
from .types import CredentialIdentifierType
from .types import GlobalCredentialIdentifier
from .types import ICredentialRepository
from .globalcredential import GlobalCredential


class BaseCredentialRepository(ICredentialRepository):
    __module__: str = 'cbra.ext.credentials'
    keychain: ApplicationKeychain

    def must_encrypt(self, credential: GlobalCredential) -> bool:
        return True

    @functools.singledispatchmethod # type: ignore
    async def get(self, credential_id: CredentialIdentifierType):
        raise TypeError(type(credential_id).__name__)
    
    async def get_global(
        self,
        credential_id: GlobalCredentialIdentifier
    ) -> GlobalCredential | None:
        raise NotImplementedError
    
    async def persist(self, credential: GlobalCredential, force: bool = False) -> None:
        dao = type(credential).parse_obj(credential.dict())
        dao.__metadata__ = credential.__metadata__
        if self.must_encrypt(credential):
            await dao.encrypt(self.keychain)
        return await self._persist(dao, force)
    
    async def persist_global(self, credential: GlobalCredential, force: bool) -> None:
        raise NotImplementedError

    @get.register # type: ignore
    async def _get_global(
        self,
        credential_id: GlobalCredentialIdentifier
    ) -> GlobalCredential | None:
        return await self.get_global(credential_id)

    @functools.singledispatchmethod
    async def _persist(self, credential: GlobalCredential, force: bool) -> None:
        raise TypeError(type(credential).__name__)
    
    @_persist.register
    async def _persist_global(self, credential: GlobalCredential, force: bool) -> None:
        return await self.persist_global(credential, force)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging

import cbra.core as cbra
from cbra.types import IDependant
from cbra.types import ISubjectResolver
from cbra.core.iam import AuthenticationService
from cbra.core.iam import AuthorizationContextFactory
from .frontendsubjectresolver import FrontendSubjectResolver
from .oidcrequestsubject import OIDCRequestSubject


class FrontendAuthorizationContextFactory(AuthorizationContextFactory, IDependant):
    """The default implementation of :class:`~cbra.types.IAuthorizationContextFactory`"""
    __module__: str = 'cbra.core.iam'
    authentication: AuthenticationService
    logger: logging.Logger = logging.getLogger('uvicorn')
    resolver: ISubjectResolver
    subject: OIDCRequestSubject

    def __init__(
        self,
        resolver: FrontendSubjectResolver = FrontendSubjectResolver.depends(),
        authentication: AuthenticationService = cbra.instance(
            name='AuthenticationService',
            missing=AuthenticationService
        )
    ):
        self.authentication = authentication
        self.resolver = resolver


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import cbra.core as cbra


class AuthenticatedEndpoint(cbra.Endpoint):

    async def get(self) -> dict[str, bool | str]:
        return {
            'remote_host': str(self.ctx.remote_host),
            'is_authenticated': self.ctx.is_authenticated()
        }


app: cbra.Application = cbra.Application()
app.inject('TRUSTED_AUTHORIZATION_SERVERS', ['https://accounts.google.com'])
app.add(AuthenticatedEndpoint)


if __name__ == '__main__':
    import uvicorn
    uvicorn.run('__main__:app', reload=True)



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from typing import Any

import pytest
from headless.ext.oauth2.types import GrantType

from cbra.ext.oauth2.models import RefreshToken
from cbra.ext.oauth2.types import RefreshTokenPolicyType
from cbra.ext.oauth2.types import RefreshTokenType



@pytest.fixture
def params() -> dict[str, Any]:
    now = datetime.now(timezone.utc)
    return {
        'created': now,
        'client_id': 'self',
        'granted': now,
        'grant_id': 0,
        'grant_type': GrantType.authorization_code,
        'ppid': 1000,
        'scope': {'email', 'openid', 'profile'},
        'sector_identifier': 'example.com',
        'sub': 1,
        'token': RefreshTokenType('foo')
    }


def test_refresh_rolling(
    params: dict[str, Any]
):
    ttl = 86400
    t1 = RefreshToken.parse_obj({
        **params,
        'renew': RefreshTokenPolicyType.rolling,
        'ttl': ttl
    })
    t2 = t1.refresh()
    assert t1.token != t2.token
    assert t2.expires == (t2.created + timedelta(seconds=ttl))
    assert not t1.is_active()


def test_refresh_static(
    params: dict[str, Any]
):
    ttl = 86400
    t1 = RefreshToken.parse_obj({
        **params,
        'renew': RefreshTokenPolicyType.static,
        'ttl': ttl
    })
    t2 = t1.refresh()
    assert t1.token != t2.token
    assert t2.expires == (t1.granted + timedelta(seconds=ttl))
    assert not t1.is_active()


def test_can_use(
    params: dict[str, Any]
):
    ttl = 86400
    t1 = RefreshToken.parse_obj({
        **params,
        'renew': RefreshTokenPolicyType.static,
        'ttl': ttl
    })
    assert t1.can_use({'openid'})
    assert not t1.can_use({'foo'})


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.types import ETagSet
from cbra.core.conf import settings
from ..secretkey import SecretKey
from .applicationkeychain import get_application_keychain
from .applicationkeychain import ApplicationKeychain
from .applicationemailsender import ApplicationEmailSender


__all__: list[str] = [
    'get_application_keychain',
    'ApplicationEmailSender',
    'ApplicationKeychain',
    'ApplicationSecretKey',
    'IfMatchRequestHeader',
    'IfMatchDestructive',
]

ApplicationSecretKey: SecretKey = SecretKey.depends()


def current_issuer(request: fastapi.Request) -> str:
    return settings.OAUTH2_ISSUER\
        or f'{request.url.scheme}://{request.url.netloc}'
CurrentIssuer: str = fastapi.Depends(current_issuer)


IfMatchDestructive: ETagSet = fastapi.Header(
    default=None,
    alias='If-Match',
    description=(
        'With the help of the `ETag` and the `If-Match` headers, '
        'you can detect mid-air edit collisions.\n\n'
        'For example, when editing a resource, the current resource content '
        'may be hashed and put into an `Etag` header in the response:\n\n'
        '```\n'
        'ETag: "33a64df551425fcc55e4d42a148795d9f25f89d4"\n'
        '```\n'
        'When saving changes to the resource the POST, PUT, PATCH or '
        'DELETE request will contain the `If-Match` header containing the '
        '`ETag` values to check freshness against:\n\n'
        '```\n'
        'If-Match: "33a64df551425fcc55e4d42a148795d9f25f89d4"\n'
        '```\n'
        'If the hashes don\'t match, it means that the resource '
        'has been modified in-between and a `412 Precondition Failed` '
        'error response is returned.'
    )
)


IfMatchRequestHeader: ETagSet = fastapi.Header(
    default=None,
    alias='If-Match',
    description=(
        'The server will only return requested resources for `GET` '
        'and `HEAD` methods, if the resource matches one of '
        'the listed `ETag` values. If the conditional does not '
        'match then the `412 (Precondition Failed)` response '
        'is returned.'
    )
)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import aorta

from cbra.types import IEndpoint
from .persister import Persister
from .resourcemodel import ResourceModel


class IResource(IEndpoint, Persister):
    model: type[ResourceModel]
    path_name: str
    publisher: aorta.types.IPublisher
    response_model_by_alias: bool = False
    response_model_exclude: set[str] | None = None
    response_model_exclude_none: bool = False
    resource_name: str
    resource_id: Any
    verbose_name: str
    verbose_name_plural: str

    async def get_object(self) -> Any:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.types import PolicyPrincipal
from cbra.ext.rbac.models import Policy


def test_add_member_to_policy():
    policy = Policy()
    policy.add("editor", "foo")
    assert policy.roles(PolicyPrincipal("foo")) == {"editor"}
    assert len(policy.members) == 1


def test_add_member_does_not_create_duplicates():
    policy = Policy()
    policy.add("editor", "foo")
    policy.add("editor", "foo")
    binding = policy.get("editor")
    assert binding is not None
    binding.update({PolicyPrincipal("foo")})
    assert len(binding.members) == 1
    assert len(policy.members) == 1


def test_merge_x_y():
    p1 = Policy()
    p1.add("editor", "foo")
    p2 = Policy()
    p2.add("owner", "foo")
    p1.merge(p2)
    assert p1.roles(PolicyPrincipal("foo")) == {"editor", "owner"}


def test_merge_existing():
    p1 = Policy()
    p1.add("editor", "foo")
    p2 = Policy()
    p2.add("editor", "bar")
    p1.merge(p2)
    assert p1.roles(PolicyPrincipal("bar")) == {"editor"}
    assert p1.roles(PolicyPrincipal("foo")) == {"editor"}


def test_get():
    p = Policy()
    p.add("foo", "foo")
    assert p.get("foo") is not None


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import fastapi
from headless.ext.oauth2 import Error
from headless.types import IClient

from .params import RequestedResourceServerClient
from .tokenhandlerendpoint import TokenHandlerEndpoint


class FrontendProxyEndpoint(TokenHandlerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    name: str = 'bff.proxy'
    path: str = '/oauth/v2/resources/{resource}{path:path}'
    resource: IClient[Any, Any] = RequestedResourceServerClient
    resource_path: str = fastapi.Path(default=..., alias='path')
    summary: str = 'Frontend Proxy Endpoint'

    async def head(self) -> fastapi.Response:
        return await self.handle('HEAD')

    async def get(self) -> fastapi.Response:
        return await self.handle('GET')

    async def post(self) -> fastapi.Response:
        return await self.handle('POST')

    async def put(self) -> fastapi.Response:
        return await self.handle('PUT')

    async def patch(self) -> fastapi.Response:
        return await self.handle('PATCH')

    async def delete(self) -> fastapi.Response:
        return await self.handle('DELETE')
    
    async def handle(self, method: str) -> fastapi.Response:
        # Discard all other headers except the Content-Type header
        content: bytes | None = None
        headers: dict[str, str | None] = {}
        if method in {'POST', 'PUT', 'PATCH'}:
            headers['Content-Type'] = self.request.headers.get('Content-Type')
            content = await self.request.body()
        try:
            response = await self.resource.request(
                content=content,
                headers={k: v for k, v in headers.items() if v},
                method=method,
                url=self.resource_path,
            )
        except Error as exception:
            return await self.on_authorization_error(exception)
        return fastapi.Response(
            content=response.content,
            headers=dict(response.headers),
            status_code=response.status_code
        )

    async def on_authorization_error(self, error: Error) -> fastapi.Response:
        # Failure during interaction with the resource server. We have no way to
        # resolve this, so instruct the client to obtain a new grant.
        return fastapi.Response(
            status_code=401,
            headers={'WWW-Authenticate': 'Bearer error="invalid_token"'}
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from ckms.core import Keychain
from ckms.types import JSONWebKeySet

from .endpoints import AuthorizationServerEndpoint
from .params import ServerKeychain


class JWKSEndpoint(AuthorizationServerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    keychain: Keychain = ServerKeychain
    name: str = 'oauth2.jwks'
    path: str = '/jwks.json'
    status_code: int = 200
    summary: str = 'JWKS Endpoint'
    tags: list[str] = ['OAuth 2.x/OpenID Connect']

    async def get(self) -> JSONWebKeySet:
        return self.keychain.as_jwks(private=False)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from collections import defaultdict

from cbra.core.conf import settings
from .staticpermissionsfinder import StaticPermissionsFinder


class SettingsPermissionsFinder(StaticPermissionsFinder): # pragma: no cover
    __module__: str = 'cbra.ext.rbac'
    available: set[str] = set()
    roles: dict[str, set[str]] = defaultdict(set)

    def __init__(self):
        super().__init__(
            available=set(getattr(settings, 'IAM_PERMISSIONS', set())),
            roles={
                role: spec['permissions']
                for role, spec in getattr(settings, 'IAM_ROLES', {}).items()
            }
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import NoReturn

from headless.ext.oauth2.models import ClaimSet

from ..params import RequestAccessToken
from ..types import BearerTokenException
from ..types import IssuedAccessToken
from .base import AuthorizationServerEndpoint


class UserInfoEndpoint(AuthorizationServerEndpoint):
    principal: RequestAccessToken # type: ignore
    name: str = 'oauth2.userinfo'
    path: str = '/userinfo'
    response_model_exclude_none: bool = True
    require_authentication: bool = False
    summary: str = 'UserInfo Endpoint'

    async def get(self) -> ClaimSet:
        if not self.principal.token_id:
            raise BearerTokenException(
                scheme='Bearer',
                error='invalid_token',
                error_description=(
                    "Provide an access token using the `Authorization` header "
                    "to query the UserInfo endpoint."
                )
            )
        token = await self.storage.fetch(self.principal.token_id)
        if token is None or token.is_expired():
            await self.on_token_invalid(token)
            assert False
        return ClaimSet.parse_obj(token.claims)
    
    async def on_token_invalid(
        self,
        token: IssuedAccessToken | None
    ) -> NoReturn:
        if token is not None:
            await self.storage.destroy(token)
        raise BearerTokenException(
            scheme='Bearer',
            error='invalid_token',
            error_description=(
                "The access token provided is expired, revoked, malformed, "
                "or invalid for other reasons."
            )
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
from typing import Literal

import pydantic
from canonical import ISO3166Alpha2
from canonical import TimezoneInfo


class SelfAssertedClaims(pydantic.BaseModel):
    cor: ISO3166Alpha2 | None = pydantic.Field(
        default=None,
        title="Country of Residence",
        max_length=2,
        description="The country an End-User is currently residing in."
    )

    initials: str | None = pydantic.Field(
        default=None,
        title="Initials",
        max_length=16,
        description="The initials of the End-User."
    )

    given_name: str | None = pydantic.Field(
        default=None,
        title="Given name",
        max_length=64,
        description=(
            "Given name(s) or first name(s) of the End-User. Note that in some "
            "cultures, people can have multiple given names; all can be present, "
            "with the names being separated by space characters."
        )
    )

    family_name: str | None = pydantic.Field(
        default=None,
        title="Family name",
        max_length=64,
        description=(
            "Surname(s) or last name(s) of the End-User. Note that in some cultures, "
            "people can have multiple family names or no family name; all can be "
            "present, with the names being separated by space characters."
        )
    )

    name_order: Literal['western', 'eastern'] | None = pydantic.Field(
        default=None,
        title="Name order",
        description="The order in which the name of the End-User is written."
    )

    middle_name: str | None = pydantic.Field(
        default=None,
        title="Middle name",
        max_length=64,
        description=(
            "Middle name(s) of the End-User. Note that in some cultures, people can "
            "have multiple middle names; all can be present, with the names being "
            "separated by space characters. Also note that in some cultures, middle "
            "names are not used."
        )
    )

    nickname: str | None = pydantic.Field(
        default=None,
        title="Nickname",
        max_length=64,
        description=(
            "Casual name of the End-User that may or may not be the same as "
            "the `given_name`. For instance, a `nickname` value of `Mike` "
            "might be returned alongside a `given_name` value of `Michael`."
        )
    )

    gender: Literal['male', 'female', 'other'] | None = pydantic.Field(
        default=None,
        title="Gender",
        description=(
            "End-User's gender. Values defined by this specification are "
            "`female` and `male`. Other values MAY be used when neither "
            "of the defined values are applicable."
        )
    )

    birthdate: datetime.date | str | None = pydantic.Field(
        default=None,
        title="Birthdate",
        max_length=8,
        description=(
            "End-User's birthday, represented as an ISO 8601:2004 `YYYY-MM-DD` "
            "format. The year MAY be `0000`, indicating that it is omitted. "
            "To represent only the year, `YYYY` format is allowed. Note that "
            "depending on the underlying platform's date related function, "
            "providing just year can result in varying month and day, so the "
            "implementers need to take this factor into account to correctly "
            "process the dates."
        )
    )

    zoneinfo: TimezoneInfo | None = pydantic.Field(
        default=None,
        title="Timezone",
        max_length=64,
        description=(
            "String from zoneinfo time zone database representing the "
            "End-User's time zone. For example, `Europe/Paris` or "
            "`America/Los_Angeles`."
        )
    )

    locale: str | None = pydantic.Field(
        default=None,
        title="Locale",
        max_length=32,
        description=(
            "End-User's locale, represented as a BCP47 language tag. This is "
            "typically an ISO 639-1 Alpha-2 language code in lowercase and "
            "an ISO 3166-1 Alpha-2 country code in uppercase, separated by "
            "a dash. For example, `en-US` or `fr-CA`. As a compatibility "
            "note, some implementations have used an underscore as the "
            "separator rather than a dash, for example, `en_US`; "
            "Relying Parties MAY choose to accept this locale syntax as well."
        )
    )


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from inspect import Parameter
from typing import Any

import fastapi
import pydantic

from cbra.types import IEndpoint
from .iresource import IResource
from .resourceaction import ResourceAction


class DetailAction(ResourceAction):
    use_resource_model: bool = True

    def get_url_pattern(
        self,
        prefix: str | None,
        endpoint: type[IResource] | None = None
    ) -> str:
        endpoint = endpoint or self.endpoint
        path = super().get_url_pattern(prefix, endpoint)
        return f'{path}/{{{str.lower(endpoint.resource_name)}_id}}'

    def is_detail(self) -> bool:
        return True

    def preprocess_parameter(self, p: Parameter) -> Parameter | None:
        """Hook to modify a parameter just before it is added to the
        new signature.
        """
        if p.name == self.path_parameter_name:
            return Parameter(
                kind=p.kind,
                name=p.name,
                annotation=p.annotation,
                default=fastapi.Path()
            )
        return super().preprocess_parameter(p)

    def preprocess_signature(
        self,
        parameters: dict[str, Parameter],
        return_annotation: Any
    ):
        if self.path_parameter_name not in parameters:
            parameters[self.path_parameter_name] = Parameter(
                kind=Parameter.POSITIONAL_ONLY,
                name=self.path_parameter_name
            )
        return super().preprocess_signature(parameters, return_annotation)

    async def process_response(
        self,
        endpoint: IEndpoint,
        response: fastapi.Response | pydantic.BaseModel | None
    ) -> fastapi.Response:
        if isinstance(response, pydantic.BaseModel)\
        and not isinstance(response, self.response_model)\
        and self.use_resource_model:
            response = self.response_model.parse_obj(response.dict(by_alias=endpoint.response_model_by_alias))
        return await super().process_response(endpoint, response)

    def parse_resource(self, resource: pydantic.BaseModel) -> pydantic.BaseModel:
        parser = self.get_write_model()
        return parser.parse_obj(resource.dict())


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class EmailChallengeResponse(pydantic.BaseModel):
    """This response is sent when the request body contained an
    `EmailChallengeRequest` object.
    """
    challenge_id: str = pydantic.Field(
        default=...,
        title="Challenge ID",
        alias='challengeId',
        description=(
            "A challenge identifier that the client must present when solving "
            "the challenge."
        ),
        max_length=64
    )


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Callable

from .container import Container
from .dependency import Dependency
from .dependencynotsatisfied import DependencyNotSatisfied


class Requirement:
    __module__: str = 'cbra.core.ioc'
    ref: Dependency | None = None
    missing: Any
    name: str

    @property
    def factory(self) -> Callable[..., Any]:
        dep = self.ref or self.missing
        assert dep is not None, self.name
        return dep.symbol

    def __init__(self, name: str, missing: Any = NotImplemented) -> None:
        self.missing = missing
        self.name = name

    def callable(self) -> bool:
        return False

    def add_to_container(self, container: Container) -> None:
        if self.ref is not None:
            return
        try:
            self.ref = container.require(self.name)
        except DependencyNotSatisfied:
            if self.missing in {NotImplemented, None}:
                raise
            self.ref = Dependency(
                name=self.name,
                qualname='NotImplemented',
                symbol=self.missing
            )

    def __repr__(self) -> str:
        return f'Requirement({self.name})'


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging

import fastapi
from headless.ext.oauth2 import Client
from headless.ext.oauth2 import Error
from headless.ext.oauth2 import models

from cbra.types import Request
from cbra.ext.oauth2.types import FatalAuthorizationException
from ..types import FrontendException
from .applicationclient import ApplicationClient
from .frontendredirecturi import FrontendRedirectURI


__all__: list[str] = [
    'FrontendTokenResponse'
]


logger: logging.Logger = logging.getLogger('cbra.endpoint')


async def get(
    request: Request,
    client: Client = ApplicationClient,
    redirect_uri: str = FrontendRedirectURI,
    expected_nonce: str = fastapi.Cookie(
        default=...,
        title="Nonce",
        description=(
            "The expected `nonce` claim in the ID Token."
        ),
        alias='bff.nonce'
    ),
    expected_state: str | None = fastapi.Cookie(
        default=None,
        title="Expected state",
        description=(
            "The expected value of the `state` query parameter."
        ),
        alias='bff.state'
    ),
    code: str | None = fastapi.Query(
        default=None,
        title='Authorization code',
        description=(
            'The authorization code generated by the authorization server. '
            'Required if using the Authorization Code Flow and the response '
            'mode is `query`, otherwise this parameter is ignored.'
        )
    ),
    state: str | None = fastapi.Query(
        default=None,
        title='State',
        description=(
            'An opaque value used by the client to maintain state '
            'between the request and callback. The authorization '
            'server includes this value when redirecting the '
            'user-agent back to the client. If supplied, this '
            'parameter **must** equal the `state` parameter used when '
            'creating the authorization request. The `state` parameter '
            'is ignored when using JARM because it is included in the '
            'JSON Web Token supplied using the `jwt` parameter per '
            'chosen response mode.'
        )
    ),
    iss: str | None = fastapi.Query(
        default=None,
        title='Issuer',
        description=(
            'Identifies the authorization server that redirected to '
            'this endpoint, as defined in RFC 9207. If the client supports '
            'OAuth 2.0 Authorization Server Issuer Identification, then '
            'this parameter is **required**, if the response mode is not '
            'JWT Secured Authorization Response Mode (JARM) per RFC 9101.\n\n'
            'The `iss` parameter value is the issuer identifier of '
            'the authorization server that created the authorization '
            'response, as defined in RFC 8414.  Its value **must** '
            'be a URL that uses the `https` scheme without any '
            'query or fragment components.'
        )
    ),
    error: str | None = fastapi.Query(
        default=None,
        title="Error code",
        description=(
            "The error code returned by the authorization server if "
            "the user cancelled the request, refused consent, or "
            "failed to authenticate."
        )
    )
) -> models.TokenResponse:
    if error:
        return None # type: ignore
    if code is None:
        raise FatalAuthorizationException(
            "The 'code' parameter is missing from the downstream "
            "response."
        )
    if state is None:
        raise FatalAuthorizationException(
            "The 'state' parameter is missing from the downstream "
            "response."
        )
    if state != expected_state:
        raise FatalAuthorizationException(
            "The 'state' parameter does not match the expected "
            "value."
        )
    if iss is not None and iss != client.issuer:
        raise FatalAuthorizationException(
            "The 'iss' parameter mismatches the expected issuer."
        )
    dto = models.AuthorizationCode(code=code, state=state)
    try:
        token = await client.token(
            dto,
            redirect_uri=(
                f'{request.url.scheme}://'
                f'{request.url.netloc}{request.url.path}'
            ),
        )
    except Error as e:
        logger.critical(
            "Unable to obtain token (error: %s, message: %s)",
            e.error, e.error_description
        )
        raise FrontendException(redirect_uri, error='temporarily_unavailable')

    await client.verify_response(
        response=token,
        nonce=expected_nonce,
        audience={client.client_id},
        issuer=iss
    )
    return token

FrontendTokenResponse: models.TokenResponse = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

from google.cloud import secretmanager

from cbra.ext import secrets


class GoogleSecret(secrets.Secret):
    provider: Literal['google']
    project: str
    name: str
    value: str | None = None
    version: str = 'latest'

    def get_value(self) -> str:
        assert self.value is not None
        return self.value


async def load(secret: GoogleSecret) -> GoogleSecret:
    client = secretmanager.SecretManagerServiceAsyncClient()
    name = f'projects/{secret.project}/secrets/{secret.name}/versions/{secret.version}'
    response = await client.access_secret_version(request={'name': name}) # type: ignore
    secret.value = bytes.decode(response.payload.data, 'utf-8')
    return secret


secrets.register('google', load)



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.types import PersistedModel



class GreatGrandChild(PersistedModel):
    pass


class GrandChild(PersistedModel):
    children: list[GreatGrandChild]


class Child(PersistedModel):
    children: list[GrandChild] = []


class Parent(PersistedModel):
    foo: int | None = None
    bar: int = 1
    baz: str = 'baz'
    children: list[Child] = []


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


class IncomingWebhook(pydantic.BaseModel):
    """Maintains all information needed to create, re-create webhooks
    with external services and verify the incoming messages.
    """
    id: str

    #: The name of the service sending the webhook messages.
    sender: str

    #: If the service allows creation of webhooks through an
    #: HTTP API, the URL of the corresponding webhook resource.
    url: str | None = None


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from datetime import timedelta
from typing import Any

from headless.ext.oauth2 import Client
from headless.ext.oauth2 import RefreshTokenException
from headless.types import IRequest

from cbra.types import ServiceNotAvailable
from cbra.ext.oauth2.client import ManagedGrant
from ..types import IPersistableCredential
from .basecredential import BaseCredential


class OpenAuthorizationAccessToken(BaseCredential):
    __module__: str = 'cbra.ext.credentials.http'
    client: Client
    credential: IPersistableCredential
    expires: datetime | None = None
    grant: ManagedGrant
    resource: str | None = None
    token: str | None = None

    def __init__(
        self,
        client: Client,
        credential: IPersistableCredential,
        *,
        resource: str | None = None,
        token: str | None = None,
        expires: datetime | None = None
    ):
        assert isinstance(credential.credential, ManagedGrant)
        self.client = client
        self.credential = credential
        self.expires = expires
        self.grant = credential.credential
        self.resource = resource
        self.token = token

    async def add_to_request(self, request: IRequest[Any]) -> None:
        now = datetime.now()
        if self.token is None or (self.expires and self.expires <= now):
            try:
                await self.refresh(now)
            except RefreshTokenException:
                self.logger.critical(
                    "Unable to obtain access token (iss: %s, client: %s)",
                    self.client.get_issuer(), self.client.client_id
                )
                raise ServiceNotAvailable
        assert self.token is not None
        request.add_header('Authorization', f'Bearer {self.token}')

    async def refresh(self, now: datetime) -> None:
        self.logger.info(
            "Refreshing access token (client: %s, sub: %s, email: %s)",
            self.client.client_id, self.grant.sub, self.grant.email
        )
        response = await self.grant.refresh(self.client)
        await self.credential.persist()
        assert self.grant.refresh_token == response.refresh_token

        # TODO: This assumes that the authorization server does not lie about
        # the lifetime of the access token. If it does, this might cause an
        # infinite loop.
        self.expires = now + timedelta(seconds=max(0, response.expires_in - 60))
        self.token = str(response.access_token)


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import asyncio
import copy
import logging.config
from typing import Any
from typing import Awaitable
from typing import Callable

import fastapi
from canonical.exc import UpstreamServiceFailure
from fastapi import FastAPI
from fastapi.params import Depends

from cbra.types import Abortable
from cbra.types import NullEmailSender
from .apiroute import APIRoute
from .conf import settings
from .endpointprovider import EndpointProvider
from .ioc import Requirement
from .localmessagetransport import LocalMessageTransport
from .messagepublisher import MessagePublisher
from .utils import parent_signature


class Application(FastAPI, EndpointProvider):
    __module__: str = 'cbra'
    _injectables: tuple[type, ...] = (
        Depends,
        Requirement
    )

    @parent_signature(FastAPI.__init__)
    def __init__(self, **kwargs: Any):
        EndpointProvider.__init__(self)
        handlers: dict[type, Any] = kwargs.setdefault('exception_handlers', {})
        handlers[Abortable] = self.on_aborted

        kwargs.setdefault('root_path', settings.ASGI_ROOT_PATH)
        self.inject('MessagePublisher', MessagePublisher)
        if not self.container.has('MessageTransport'):
            self.inject('MessageTransport', LocalMessageTransport)
        if not self.container.has('EmailSender'):
            self.container.provide('EmailSender', {
                'qualname': '_',
                'symbol': NullEmailSender
            })
        FastAPI.__init__(self, **kwargs)
        self.router.route_class = APIRoute
        self.add_event_handler('startup', self.setup_logging) # type: ignore
        if settings.DEBUG_RESPONSE_TIME is not None:
            self.middleware('http')(self.delay_response)

        self.add_exception_handler( # type: ignore
            UpstreamServiceFailure,
            self.on_upstream_service_failure
        )

    async def on_upstream_service_failure(
        self,
        request: fastapi.Request,
        exc: UpstreamServiceFailure
    ) -> fastapi.Response:
        """Invoked when an upstream service failed."""
        return fastapi.Response(status_code=503)

    def inject(self, name: str, value: Any) -> None:
        """Inject a value into the dependencies container."""
        self.container.inject(name, value)

    def logging_config(self):
        config = copy.deepcopy(settings.LOGGING)
        if not settings.DEBUG:
            # Remove console handler when not running in debug mode and its
            # not explicitely enabled in the settings.
            config['handlers']['console'] = {'class': 'logging.NullHandler'}
        return config

    def setup_logging(self) -> None:
        logging.config.dictConfig(self.logging_config())

    async def on_aborted(
        self,
        request: fastapi.Request,
        exc: Abortable
    ) -> fastapi.Response:
        return await exc.as_response()

    @parent_signature(FastAPI.add_api_route)
    def add_api_route(
        self,
        endpoint: Callable[..., Any],
        *args: Any,
        **kwargs: Any
    ) -> None:
        self.update_requirements(endpoint)
        return super().add_api_route(endpoint=endpoint, *args, **kwargs)

    @parent_signature(FastAPI.head)
    def head(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.head, *a, **k)

    @parent_signature(FastAPI.get)
    def get(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.get, *a, **k)

    @parent_signature(FastAPI.post)
    def post(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.post, *a, **k)

    @parent_signature(FastAPI.patch)
    def patch(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.patch, *a, **k)

    @parent_signature(FastAPI.put)
    def put(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.put, *a, **k)

    @parent_signature(FastAPI.trace)
    def trace(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.trace, *a, **k)

    @parent_signature(FastAPI.options)
    def options(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.options, *a, **k)

    @parent_signature(FastAPI.delete)
    def delete(self, *a: Any, **k: Any):
        return self.discover_requirements(FastAPI.delete, *a, **k)
    
    async def delay_response(
        self,
        request: fastapi.Request,
        call_next: Callable[[fastapi.Request], Awaitable[fastapi.Response]]
    ) -> fastapi.Response:
        if isinstance(settings.DEBUG_RESPONSE_TIME, int):
            await asyncio.sleep(settings.DEBUG_RESPONSE_TIME/1000)
        return await call_next(request)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import NoReturn

import fastapi
from canonical import EmailAddress
from headless.ext.oauth2 import Client
from headless.ext.oauth2.models import OIDCToken
from headless.ext.oauth2.models import TokenResponse

from cbra.core import Endpoint
from ...params import Error
from ...params import ClientStorage
from ...types import AuthorizationState
from ...types import AuthorizeResponse
from ...types import FatalAuthorizationException
from ...types import IFrontendStorage
from ...types import OIDCProvider
from ..models import ManagedGrant


class BaseCallbackEndpoint(Endpoint):
    __module__: str = 'cbra.ext.oauth2.server'
    error: Error | None = Error.depends()
    expects_id_token: bool = True
    params: AuthorizeResponse
    provider: OIDCProvider
    token: TokenResponse | None
    state: AuthorizationState
    storage: IFrontendStorage = ClientStorage
    verify_email: bool = True

    def get_redirect_uri(self) -> str:
        assert self.state is not None
        return self.state.redirect_uri

    def is_trusted_email(self, email: EmailAddress | None) -> bool:
        if not self.verify_email or not email:
            return False
        return self.provider.is_trusted_email(email)

    async def authorize_grant(
        self,
        client: Client,
        params: AuthorizationState,
        response: TokenResponse,
        claims: OIDCToken | None,
    ) -> ManagedGrant:
        """Authorizes the grant received from the authorization server."""
        raise NotImplementedError

    async def post(self) -> fastapi.Response:
        """Callback endpoint for responses from OAuth 2.x/OpenID Connect authorization
        servers. A `POST` request is used when the choses response mode produces such
        a request from the authorization server, such as `form_post` or `form_post.jwt`.
        """
        raise NotImplementedError

    async def get(self) -> fastapi.Response:
        """Callback endpoint for responses from OAuth 2.x/OpenID Connect authorization
        servers.
        """
        await self.session
        try:
            return await self.handle(self.params)
        except Exception as exception:
            response = await self.on_fatal_exception(exception)
            if response:
                return response
            raise

    async def get_client(self) -> Client:
        """Return a :class:`headless.ext.oauth2.Client` instance that can exchange
        the authorization code.
        """
        raise NotImplementedError

    async def on_fatal_exception(self, exception: Exception) -> fastapi.Response | None:
        raise exception
    
    async def on_malformed_id_token(self, exception: Exception) -> NoReturn:
        raise exception
    
    async def on_success(
        self,
        client: Client,
        params: AuthorizationState,
        response: TokenResponse,
        claims: OIDCToken | None,
        grant: None = None
    ) -> fastapi.Response:
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=params.return_uri
        )
    
    async def handle(self, authorization: AuthorizeResponse) -> fastapi.Response:
        """Handles the response from the authorization server."""
        async with await self.get_client() as client:
            params = await authorization.verify(client, self.storage)
            token = await authorization.obtain(client, params)
            return await self.handle_token_response(client, params, token)

    async def handle_token_response(
        self,
        client: Client,
        params: AuthorizationState,
        response: TokenResponse
    ) -> fastapi.Response:
        claims = None
        if self.expects_id_token:
            if not response.id_token:
                raise FatalAuthorizationException(
                    "The downstream authorization server did not include "
                    "an OIDC ID Token in the response."
                )
            try:
                claims = response.id_token.parse()
            except Exception as exception:
                # Parse error, but should be more specifically declared (TODO).
                self.logger.exception('Caught fatal %s', type(exception).__name__)
                await self.on_malformed_id_token(exception)
            claims.email_verified = self.is_trusted_email(claims.email)
            self.logger.debug("Receiving OIDC token from %s", claims.iss)
        grant = await self.authorize_grant(client, params, response, claims)
        return await self.on_success(
            client, params, response, claims,
            await self.persist(client, params, response, claims, grant)
        )
        
    async def persist(
        self,
        client: Client,
        params: AuthorizationState,
        response: TokenResponse,
        claims: OIDCToken | None,
        grant: ManagedGrant
    ) -> None:
        pass


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
import hmac
import json
from typing import Any
from typing import AsyncIterable
from typing import Callable
from typing import Coroutine
from typing import TypeAlias

import fastapi
import pydantic
import pytest
import pytest_asyncio
from headless.core import httpx

import cbra.core as cbra
from cbra.ext import webhooks
from cbra.types import IVerifier
from cbra.types import Request


RequestFactory: TypeAlias = Callable[[str, dict[str, Any]], Coroutine[Any, Any, webhooks.WebhookResponse]]
SECRET_KEY: bytes = b'test'


class BookWebhookEnvelope(webhooks.WebhookEnvelope):
    content: dict[str, Any]
    signature: bytes | None

    def __init__(
        self,
        content: dict[str, Any],
        signature: bytes | None = fastapi.Header(None, alias='X-Webhook-Signature'),
        topic: str = fastapi.Header(..., alias='X-Webhook-Topic'),
    ) -> None:
        self.content = content
        self.event_name = topic
        self.signature = signature
        if signature is not None:
            self.signature = base64.b64decode(signature)

    def get_message(self) -> pydantic.BaseModel:
        return BookWebhookMessage.parse_obj(self.content)
    
    async def verify(self, request: Request, verifier: IVerifier) -> bool:
        if not self.signature:
            return False
        return await verifier.verify(self.signature, await request.body())


class BookWebhookMessage(pydantic.BaseModel):
    foo: int


class BookWebhookEndpoint(webhooks.WebhookEndpoint):
    envelope: BookWebhookEnvelope
    domain: str = 'cbra.localhost'

    async def on_hello(self, message: BookWebhookMessage) -> None:
        assert isinstance(message, BookWebhookMessage)

    async def on_exception(self, message: BookWebhookMessage) -> None:
        raise Exception


class AuthenticatedBookWebhookEndpoint(BookWebhookEndpoint):
    signed: bool = True

    async def get_verifier(self) -> IVerifier:
        return webhooks.HMACSHA256WebhookVerifier(SECRET_KEY)


@pytest.fixture # type: ignore
def app() -> cbra.Application:
    app = cbra.Application()
    app.add(BookWebhookEndpoint, path='/')
    return app


@pytest.fixture
def secret_key() -> bytes:
    return SECRET_KEY


@pytest_asyncio.fixture # type: ignore
async def client(app: cbra.Application) -> AsyncIterable[httpx.Client]:
    params: dict[str, Any] = {
        'app': app,
        'base_url': "https://cbra"
    }
    async with httpx.Client(**params) as client:
        yield client


@pytest.fixture
def request_factory(
    client: httpx.Client,
    secret_key: bytes
) -> RequestFactory:
    async def f(* ,topic: str, msg: dict[str, Any], sign: bool = False, **kwargs: Any) -> webhooks.WebhookResponse:
        content = str.encode(json.dumps(msg), 'utf-8')
        digest = hmac.digest(
            secret_key,
            content,
            'sha256'
        )
        headers = kwargs.setdefault('headers', {})
        headers['Content-Type'] = "application/json"
        headers['X-Webhook-Topic'] = topic
        if sign:
            headers['X-Webhook-Signature'] = base64.b64encode(digest)
        response = await client.post(content=content, **kwargs)
        if response.status_code != 200:
            raise Exception(response.status_code, response.content)
        return webhooks.WebhookResponse.parse_obj(await response.json())
    return f # type: ignore


@pytest.fixture
def sign() -> bool:
    return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic
from canonical import EmailAddress

from .principalmodel import PrincipalModel


class EmailPrincipal(PrincipalModel):
    kind: Literal['EmailAddress'] 
    email: EmailAddress = pydantic.Field(..., primary_key=True)

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        value: EmailAddress | None = values.pop('principal', None)
        if isinstance(value, EmailAddress):
            values['email'] = value
        return values

    def __str__(self) -> str:
        return str(self.email)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import TypeAlias
from typing import Union

from .applicationclient import ApplicationClient
from .authorizationcodegrant import AuthorizationCodeGrant
from .authorizationrequest import AuthorizationRequest
from .authorizationrequestclient import AuthorizationRequestClient
from .authorizationrequestparameters import AuthorizationRequestParameters
from .authorizationstate import AuthorizationState
from .beginonboardrequest import BeginOnboardRequest
from .beginonboardresponse import BeginOnboardResponse
from .client import Client
from .clientcredentialsgrant import ClientCredentialsGrant
from .currentauthorizationrequest import CurrentAuthorizationRequest
from .externalauthorizationstate import ExternalAuthorizationState
from .frontendloginrequest import FrontendLoginRequest
from .frontendloginresponse import FrontendLoginResponse
from .grant import Grant
from .patchauthorizationrequest import PatchAuthorizationRequest
from .patchauthorizationresponse import PatchAuthorizationResponse
from .resourceowner import ResourceOwner
from .refreshtoken import RefreshToken
from .refreshtokengrant import RefreshTokenGrant
from .sector import Sector
from .managedgrant import ManagedGrant


__all__: list[str] = [
    'ApplicationClient',
    'AuthorizationCodeGrant',
    'AuthorizationRequest',
    'AuthorizationRequestClient',
    'AuthorizationRequestParameters',
    'AuthorizationState',
    'BeginOnboardRequest',
    'BeginOnboardResponse',
    'ClientCredentialsGrant',
    'CurrentAuthorizationRequest',
    'ExternalAuthorizationState',
    'FrontendLoginRequest',
    'FrontendLoginResponse',
    'Grant',
    'ManagedGrant',
    'PatchAuthorizationRequest',
    'PatchAuthorizationResponse',
    'RefreshToken',
    'RefreshTokenGrant',
    'ResourceOwner',
    'Sector',
]


AuthorizationServerModel: TypeAlias = Union[
    AuthorizationRequest,
    AuthorizationRequestParameters,
    AuthorizationState,
    Client,
    Sector
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from headless.core import BaseCredential
from headless.types import IRequest

from .iaccesstoken import IAccessToken
from .iaccesstokenobtainer import IAccessTokenObtainer


class BearerTokenCredential(BaseCredential):
    __module__: str = 'cbra.ext.oauth2.types'
    access_token: IAccessToken | None
    factory: IAccessTokenObtainer

    def __init__(
        self,
        factory: IAccessTokenObtainer,
        access_token: IAccessToken | None = None,
    ):
        self.access_token = access_token
        self.factory = factory

    async def add_to_request(self, request: IRequest[Any]) -> None:
        if self.access_token is None or self.access_token.is_expired():
            self.access_token = await self.factory.obtain()
        request.add_header('Authorization', f'Bearer {str(self.access_token)}')


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.


class ValidationError(ValueError):
    """Indicares the validation error on a single scalar or composite
    data element.
    """
    __module__: str = 'cbra.types'
    name: str
    codes: list[str]

    def __init__(self, name: str, codes: list[str]) -> None:
        self.name = name
        self.codes = codes


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .credentialscope import CredentialScope
from .globalcredentialidentifier import GlobalCredentialIdentifier
from .icredentialrepository import CredentialIdentifierType
from .icredentialrepository import ICredentialRepository
from .ipersistablecredential import IPersistableCredential


__all__: list[str] = [
    'CredentialIdentifierType',
    'CredentialScope',
    'GlobalCredentialIdentifier',
    'ICredentialRepository',
    'IPersistableCredential',
]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
from collections import OrderedDict
from typing import Any
from typing import Callable
from typing import TypeVar


F = TypeVar('F', bound=Callable[..., Any])


class InheritDependenciesDecorator:
    parent: type[Any]
    signature: inspect.Signature

    def __init__(
        self,
        parent: type[Any]
    ):
        self.parent = parent
        self.signature = inspect.signature(parent)

    def __call__(
        self,
        func: F
    ) -> F:
        sig = inspect.signature(func)
        params: OrderedDict[str, inspect.Parameter] = OrderedDict([
            (name, param)
            for name, param in sig.parameters.items()
            if name not in ('args', 'kwargs')
        ])

        # Add the parameters of the parent, if it is not overriden
        # in the signature of func.
        for name, param in self.signature.parameters.items():
            if name in params or name in {'args', 'kwargs'}:
                continue
            params[name] = param
        func.__signature__ = sig.replace( # type: ignore
            parameters=list(params.values())
        )
        return func


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol
from typing import TypeAlias

from .globalcredentialidentifier import GlobalCredentialIdentifier


__all__: list[str] = [
    'CredentialIdentifierType',
    'ICredentialRepository'
]


CredentialIdentifierType: TypeAlias = GlobalCredentialIdentifier


class ICredentialRepository(Protocol):
    __module__: str = 'cbra.ext.credentials.types'

    async def get(self, __credential_id: CredentialIdentifierType):
        ...

    async def persist(self, credential: Any, force: bool = False) -> None:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from headless.ext.oauth2.models import OIDCToken

from cbra.core.conf import settings
from cbra.types import NotFound
from cbra.types import SessionRequestPrincipal
from .tokenhandlerendpoint import TokenHandlerEndpoint
from .types import OIDCTokenSubjectIdentifier


class FrontendUserEndpoint(TokenHandlerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    name: str = 'bff.user'
    path: str = '/oauth/v2/user'
    principal: SessionRequestPrincipal # type: ignore
    response_model_exclude: set[str] = {
        "aud",
        "azp",
        "at_hash",
        "c_hash",
        "exp",
        "iat",
        "iss",
        "nbf",
        "nonce",
    }
    response_model_exclude_none: bool = True
    summary: str = "Current User"

    async def get(self) -> OIDCToken:
        await self.session
        if not self.session.subject:
            self.logger.debug("User is not authenticated")
            raise NotFound
        oidc = await self.storage.get(OIDCTokenSubjectIdentifier(self.session.subject.sha256))
        if oidc is None:
            self.logger.debug("No OIDC token found for authenticated user.")
            raise NotFound
        if oidc.email:
            oidc.email_verified = settings.APP_ISSUER_TRUST # TODO
        return oidc
    
    async def delete(self) -> None:
        await self.session
        self.session.logout()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cbra.core.iam.types import Subject
from cbra.types import Forbidden
from .endpoints import AuthorizationServerEndpoint
from .server.endpoints.models import CurrentSubjectResponse
from .server.endpoints.models import SelfAssertedClaimsRequest


class CurrentSubjectEndpoint(AuthorizationServerEndpoint):
    name: str = 'oauth2.me'
    path: str = '/me'
    summary: str = 'Current User Endpoint'

    async def get(self) -> CurrentSubjectResponse:
        await self.session
        claims: dict[str, Any] = {
            'sub': 'allUsers',
            'mailboxes': []
        }
        if self.is_authenticated():
            claims['sub'] = 'self'
            subject = await self.get_subject()
            for principal in subject.get_principals():
                if principal.spec.kind != 'EmailAddress':
                    continue
                claims['mailboxes'].append({
                    'principal_id': principal.key,
                    'email': principal.spec.email
                })
            claims.update(subject.get_claims())

        return CurrentSubjectResponse.parse_obj(claims)
    
    async def patch(self, dto: SelfAssertedClaimsRequest) -> None:
        """Self-assert claims regarding the identity of the Subject."""
        await self.session
        if not self.is_authenticated():
            raise Forbidden
        await self.on_claims_asserted(await self.get_subject(), dto)

    async def on_claims_asserted(self, subject: Subject, dto: Any) -> None:
        self.logger.critical(
            f'{type(self).__name__}.on_claims_asserted() is not implemented.'
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from cbra.types import RequestLanguage


def test_multiple_weighted():
    l = RequestLanguage("nl-NL,nl;q=0.9,en-US;q=0.8,en;q=0.7,es;q=0.6")
    assert l.languages[0].code == 'nl-NL'
    assert l.languages[1].code == 'nl'
    assert l.languages[2].code == 'en-US'
    assert l.languages[3].code == 'en'
    assert l.languages[4].code == 'es'


@pytest.mark.parametrize("language", [
    "nl-nl",
    "nl",
    "en-us",
    "en",
    "es"
])
def test_selectable(language: str):
    l = RequestLanguage("nl-NL,nl;q=0.9,en-US;q=0.8,en;q=0.7,es;q=0.6")
    assert str.lower(l.select({language})) == str.lower(language)


def test_no_header_provided_returns_default_has_en_default():
    l = RequestLanguage()
    assert l.select({'foo', 'bar', 'baz'}) == 'en'


def test_none_provided_returns_default_has_en_default():
    l = RequestLanguage(None)
    assert l.select({'foo', 'bar', 'baz'}) == 'en'


def test_none_provided_returns_default():
    l = RequestLanguage(None)
    assert l.select({'foo', 'bar', 'baz'}, 'nl') == 'nl'


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime
import secrets

import pydantic
import uvicorn

import cbra.core as cbra


class BookPublication(pydantic.BaseModel):
    published: datetime.date
    country_code: str


class Book(cbra.ResourceModel):
    id: int | None = cbra.Field(
        default=None,
        read_only=True,
        path_alias='book_id',
        primary_key=True
    )
    title: str
    publications: list[BookPublication]


class BookResource(cbra.Resource, cbra.Create, cbra.Delete, cbra.Retrieve, cbra.Update, model=Book):
    books: dict[int, Book] = {
        1: Book(
            id=1,
            title="The Hitchhiker's Guide to the Galaxy",
            publications=[
                BookPublication(country_code='UK', published=datetime.date(1979, 10, 12)),
                BookPublication(country_code='US', published=datetime.date(1980, 10, 1)),
            ]
        )
    }

    async def can_create(self, resource: Book) -> bool:
        return not any([x.title == resource.title for x in self.books.values()])

    async def delete(self, resource: Book):
        assert resource.id is not None
        self.books.pop(resource.id)

    async def get_object(self) -> Book | None:
        return self.books.get(int(self.request.path_params['book_id']))

    async def persist(self, resource: Book, create: bool = False) -> Book:
        if create:
            assert resource.id is None
            resource.id = secrets.choice(range(1000, 9999))
        assert resource.id is not None
        self.books[resource.id] = resource
        return resource


app = cbra.Application(docs_url='/ui')
BookResource.add_to_router(app, path='/')

if __name__ == '__main__':
    uvicorn.run('__main__:app', reload=True) # type: ignore



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cbra.ext.credentials import BaseCredentialRepository
from cbra.ext.credentials import GlobalCredential
from cbra.ext.credentials import GlobalCredential
from cbra.ext.credentials.types import GlobalCredentialIdentifier
from ...basemodelrepository import BaseModelRepository


class DatastoreCredentialRepository(BaseModelRepository, BaseCredentialRepository):
    __module__: str = 'cbra.ext.google.impl.credentials'

    async def get_global(
        self,
        credential_id: GlobalCredentialIdentifier
    ) -> GlobalCredential | None:
        e = await self.get_entity_by_key(self.model_key(GlobalCredential, credential_id))
        c = None
        if e is not None:
            c = self.restore(GlobalCredential, e)
            c.__repository__ = self
            await c.decrypt(self.keychain)
        return c

    async def persist_global(self, credential: GlobalCredential, force: bool) -> None:
        await self.persist_model(credential, force=force)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from cbra.ext.rbac import StaticPermissionsFinder


@pytest.mark.asyncio
async def test_get_single():
    finder = StaticPermissionsFinder(
        available={'foo'},
        roles={'viewer': {'foo'}}
    )
    assert {'foo'} == await finder.get({'viewer'})


@pytest.mark.parametrize("available", [
    {'bar.get', 'bar.list', 'foo.get', 'foo.list'}
])
@pytest.mark.parametrize("roles,permissions", [
    ({"viewer"}, {"foo.get", "foo.list"}),
    ({"editor"}, {"foo.create", "foo.delete", "foo.get", "foo.list", "foo.update"}),
    ({"anonymous"}, set()), # type: ignore
    ({"role1", "role2"}, {"foo.get", "foo.list"}),
    ({"owner"}, {"bar.get", "bar.list", "foo.get", "foo.list"}),
    ({"wildcard"}, {"foo.get", "bar.get"}),
])
@pytest.mark.asyncio
async def test_get(available: set[str], roles: set[str], permissions: set[str]):
    finder = StaticPermissionsFinder(
        available=available,
        roles={
            'viewer': {
                'foo.get',
                'foo.list',
            },
            'editor': {
                'foo.create',
                'foo.delete',
                'foo.get',
                'foo.list',
                'foo.update',
            },
            'role1': {
                "foo.get",
            },
            'role2': {
                "foo.list",
            },
            'owner': {'*'},
            'wildcard': {'*.get'}
        }
    )
    assert permissions == await finder.get(roles)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime
from typing import Literal

import pydantic

from cbra.types import PersistedModel
from ..types import SubjectLifecycleType


class BaseSubject(PersistedModel):
    kind: Literal['User']
    uid: int | None = pydantic.Field(
        default=None,
        auto_assign=True
    )
    created: datetime
    seen: datetime
    status: SubjectLifecycleType = SubjectLifecycleType.pending

    def is_encypted(self) -> bool:
        raise NotImplementedError


# Copyright (C) 2020-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`MessagePublished`."""
import pydantic

from .pubsubmessage import PubsubMessage

SUBSCRIPTION_NAME_REGEX = r'^projects/[\-a-z0-9]{6,30}/subscriptions/.*$'


class MessagePublished(pydantic.BaseModel):
    """A datastructure containing a message that was published to a Google Pub/Sub
    topic and received by the server through a subscription.
    """
    subscription: str = pydantic.Field(
        default=...,
        title="Subscription",
        description=(
            "The subscription through which the message was delivered to the "
            "endpoint."
        ),
        regex=SUBSCRIPTION_NAME_REGEX,
        example="projects/myproject/subscriptions/mysub"
    )

    message: PubsubMessage



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib
import hmac

import fastapi


class Request(fastapi.Request):
    __module__: str = 'cbra.core'

    @property
    def sha256(self) -> bytes:
        if not hasattr(self, "_body"):
            raise ValueError("Body not available")
        return hashlib.sha256(self._body).digest()
    
    def hmac(self, secret: bytes, algorithm: str = 'sha256') -> bytes:
        assert self._body is not None
        return hmac.digest(secret, self._body, algorithm)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse

import fastapi

from cbra.core.conf import settings
from cbra.types import Forbidden
from cbra.types import NotFound
from .params import ClientProvider
from .params import LocalClientProvider
from .endpoints import AuthorizationServerEndpoint
from .models import ExternalAuthorizationState
from .models import BeginOnboardRequest
from .models import BeginOnboardResponse


class OnboardingEndpoint(AuthorizationServerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    name: str = 'oauth2.oidc'
    status_code: int = 200
    summary: str = 'OIDC Onboarding Endpoint'
    path: str = '/oidc/{client_id}'
    provider: ClientProvider = LocalClientProvider

    async def post(
        self,
        dto: BeginOnboardRequest,
        client_id: str = fastapi.Path(
            default=...,
            title="Client ID",
            description=(
                "Specifies the OIDC client application used to "
                "authenticate with the downstream authorization "
                "server."
            )
        )
    ) -> BeginOnboardResponse:
        """Authenticate and register a new user with an external (downstream)
        OIDC-capable authorization server. Redirect the user-agent to the
        Authorization Endpoint of the downstream server.

        The authorization server is pre-registered with this server under the
        `client_id` provided in the path. Consult the documentation or
        system administrator(s) for the list of available clients.
        """
        if not self.can_redirect(dto.next):
            raise Forbidden
        provider = await self.provider.get(client_id)
        if provider is None:
            raise NotFound
        async with provider:
            state = ExternalAuthorizationState.new(
                provider, client_id, dto.ctx, dto.request, dto.next
            )
            redirect_uri = await state.get_redirect_uri(
                provider=provider,
                callback_uri=str(self.request.url_for('oauth2.callback')),
            )
        await self.storage.persist(state)
        return BeginOnboardResponse(redirect_url=redirect_uri)

    def can_redirect(self, redirect_uri: str) -> bool:
        """Return a boolean if the user-agent may request a redirect to the
        given URI.
        """
        p = urllib.parse.urlparse(redirect_uri)
        return any([
            all([
                p.netloc == self.request.url.netloc,
                p.scheme == self.request.url.scheme,
            ]),
            p.netloc in settings.LOGIN_AUTHORIZED_DOMAINS
        ])


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from datetime import datetime
from datetime import timezone

import pydantic
from canonical import EmailAddress
from headless.ext import oauth2

from cbra.types import VersionedCipherText
from cbra.ext.security import ApplicationKeychain
from ...types import NoRefreshTokenReturned


class ManagedGrant(pydantic.BaseModel):
    """A grant received from an external identity
    provider.
    """
    client_id: str
    email: EmailAddress | None = None
    iss: str
    received: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    refresh_token: str | VersionedCipherText
    scope: set[str] = set()
    sub: str | None = None
    token_endpoint: str
    version: int = 0

    @classmethod
    def parse_response(
        cls,
        client: oauth2.Client,
        response: oauth2.TokenResponse,
        sub: str | None = None
    ):
        """Create a new :class:`ManagedGrant` using the response from the
        Token Endpoint.
        """
        return cls.parse_obj({
            'client_id': client.client_id,
            'iss': client.get_issuer(),
            'refresh_token': response.refresh_token,
            'scope': set(filter(bool, re.split('\\s+', response.scope or ''))),
            'sub': sub,
            'token_endpoint': client.server.token_endpoint
        })

    async def decrypt(self, keychain: ApplicationKeychain):
        if isinstance(self.refresh_token, VersionedCipherText):
            self.refresh_token = await keychain.decrypt(
                self.refresh_token,
                parser=lambda _, pt: bytes.decode(pt, 'utf-8')
            )

    async def encrypt(self, keychain: ApplicationKeychain):
        if not isinstance(self.refresh_token, VersionedCipherText):
            self.refresh_token = await keychain.encrypt(self.refresh_token)

    async def refresh(self, client: oauth2.Client, resource: str | None = None):
        """Invoke the Token Endpoint to refresh an access token and
        refresh token, and correspondingly update the :attr`:refresh_token`
        attribute.
        """
        assert isinstance(self.refresh_token, str)
        response = await client.refresh_token(
            token=self.refresh_token,
            resource=resource,
            scope=self.scope or None
        )
        if response.refresh_token is None:
            raise NoRefreshTokenReturned(
                "The authorization server did not return a refresh "
                "token."
            )
        self.refresh_token = response.refresh_token
        return response


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from canonical import EmailAddress


class SessionClaims(pydantic.BaseModel):
    ctx: str | None = None
    uid: int | None = None
    icg: bool | None = None
    iss: str | None = None
    sub: str | None = None
    uai: str | None = None
    email: EmailAddress | None = None
    email_verified: bool = False
    auth_time: int | None = None


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
""".. _guides-authentication:

=======================
Authenticating requests
=======================
The :mod:`cbra` module comes with a basic authentication system that enables
resolving an incoming HTTP request to a subject identifier, based on the
parameters included in the request, such as headers or cookies. A subject
identifier uniquely points to an account of some kind (user account, service
account) that has access to certain protected resources.

What is authentication?
=======================
Authentication is a term that refers to the process of proving that some fact
or some document is genuine. In computer science, this term is typically
associated with proving a user's identity. The process of authentication can
thus be modelled into three models:

- A **Subject** is a person or a service that has one or many identities.
- A **Principal** represents a specific identity of a Subject, such as
  an email address, phone number or account identifier.
- A **Credential** is used to assert the ownership of or access to a
  specific Principal.

Practically speaking, *authentication* may then be defined as follows:

A **Subject** uses a combination of a **Principal** and a **Credential** to
assert its identity. These concepts are reflected in the abstract interfaces
specified by :mod:`cbra.types`, being :class:`~cbra.types.ICredential`,
:class:`~cbra.types.IRequestPrincipal` and :class:`~cbra.types.ISubject`. The default
implementation in :mod:`cbra.core.iam` supports RFC 9068 and
OpenID Connect ID Tokens to authenticate requests, but can quite easily
extended to support other mechanisms, like your own bearer token.


Configuring authentication
==========================
Configuring :mod:`cbra` with the default implementation is as simple as setting
:attr:`~cbra.core.conf.settings.TRUSTED_AUTHORIZATION_SERVERS` in your settings
module:

.. code:: python

    # settings.py
    TRUSTED_AUTHORIZATION_SERVERS: list[str] = [
        'https://accounts.google.com'
    ]


Authenticated requests
======================
The :mod:`cbra` IAM framework may be used with :class:`~cbra.core.Endpoint`
and :class:`cbra.core.Resource`, or plain functions as you might be used to
with FastAPI:

.. code:: python

    import fastapi
    import cbra.core as cbra
    from cbra.core.iam import AuthorizationContextFactory
    from cbra.types import Principal

    app: cbra.Application = cbra.Application()

    @app.get('/')
    async def f(
        request: fastapi.Request,
        factory: AuthorizationContextFactory = fastapi.Depends(),
        principal: Principal = fastapi.Depends(Principal.fromrequest)
    ):
        ctx = await factory.authenticate(request, principal)
        return {
            'remote_host': ctx.remote_host,
            'is_authenticated': ctx.is_authenticated()
        }

    if __name__ == '__main__':
        import uvicorn
        uvicorn.run(app)

Similarly, :class:`~cbra.core.Endpoint` and :class:`~cbra.core.Resource`
may be used with the authentication framework:

.. code:: python

    import cbra.core as cbra


    class AuthenticatedEndpoint(cbra.Endpoint):

        async def get(self) -> dict[str, bool | str]:
            return {
                'remote_host': str(self.ctx.remote_host),
                'is_authenticated': self.ctx.is_authenticated()
            }


    app: cbra.Application = cbra.Application()
    app.inject('TRUSTED_AUTHORIZATION_SERVERS', ['https://accounts.google.com'])
    app.add(AuthenticatedEndpoint)


    if __name__ == '__main__':
        import uvicorn
        uvicorn.run('__main__:app', reload=True)

To run the above example, the following script obtains an OpenID Connect
ID Token from Google:

.. code:: python

    import google.oauth2.id_token
    import google.auth.transport.requests

    if __name__ == '__main__':
        request = google.auth.transport.requests.Request()
        audience = 'http://localhost:8000'
        id_token = google.oauth2.id_token.fetch_id_token(request, audience)
        headers = {'Authorization': f'bearer {id_token}'}
        print(id_token)
"""
from .authenticatedcontext import AuthenticatedContext
from .authenticationservice import AuthenticationService
from .authorizationcontextfactory import AuthorizationContextFactory
from .isubjectrepository import ISubjectRepository
from .memorysubjectrepository import MemorySubjectRepository
from .nullauthorizationcontext import NullAuthorizationContext
from .nullsubjectesolver import NullSubjectResolver
from .subject import Subject as RequestSubject
from .subjectresolver import SubjectResolver
from .types import IUserOnboardingService


__all__: list[str] = [
    'AuthenticatedContext',
    'AuthenticationService',
    'AuthorizationContextFactory',
    'ISubjectRepository',
    'IUserOnboardingService',
    'MemorySubjectRepository',
    'NullAuthorizationContext',
    'NullSubjectResolver',
    'RequestSubject',
    'SubjectResolver',
]